# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Data utility functions."""
from typing import List, Union

import pandas as pd

from .._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"


def ensure_df_datetimes(
    data: pd.DataFrame,
    columns: Union[str, List[str], None] = None,
    add_utc_tz: bool = True,
) -> pd.DataFrame:
    """
    Return dataframe with converted TZ-aware timestamps.

    Parameters
    ----------
    data : pd.DataFrame
        Input dataframe
    columns : Union[str, List[str], None], optional
        column (str) or list of columns to convert, by default None.
        If this parameter is not supplied then any column containing
        the substring "time" is used as a candidate for conversion.
    add_utc_tz: bool, optional
        If True any datetime columns in the `columns` parameter (
        (or default `'.*time.*'` columns) that are timezone-naive,
        will be converted to Timezone-aware timestamps marked as UTC.

    Returns
    -------
    pd.DataFrame
        Converted DataFrame.

    """
    if not columns:
        columns = list(data.filter(regex=".*[Tt]ime.*").columns)  # type: ignore
    if isinstance(columns, str):
        columns = [columns]
    col_map = {
        col: "datetime64[ns, UTC]"
        for col in set(columns)
        if col in data.columns and not pd.api.types.is_datetime64_any_dtype(data[col])
    }
    converted_data = data.astype(col_map, errors="ignore")

    # Look for any TZ-naive columns in the list
    if add_utc_tz:
        localize_cols = {
            col for col in columns if col in data.select_dtypes("datetime")
        }
        for col in localize_cols:
            converted_data[col] = converted_data[col].dt.tz_localize(
                "UTC", ambiguous="infer", nonexistent="shift_forward"
            )
    return converted_data


def df_has_data(data) -> bool:
    """Return true if data is a pd.DataFrame and is not empty."""
    return isinstance(data, pd.DataFrame) and not data.empty



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Module for pre-defined widget layouts."""
import os

import ipywidgets as widgets
from IPython.display import display
from ipywidgets import Layout

from .._version import VERSION
from .core import IPyDisplayMixin, RegisteredWidget

__version__ = VERSION
__author__ = "Ian Hellen"


class GetEnvironmentKey(RegisteredWidget, IPyDisplayMixin):
    """
    GetEnvironmentKey.

    Tries to retrieve an environment variable value. The value
    can be changed/set and optionally saved back to the system
    environment.
    """

    _NB_PARAMS = {"value": "_value"}

    def __init__(
        self,
        env_var: str,
        help_str: str = None,
        description: str = "Enter the value: ",
        auto_display: bool = False,
        **kwargs,
    ):
        """
        Create a new instance of GetEnvironmentKey.

        Parameters
        ----------
        env_var : str
            Name of the environment variable.
        help_str : str, optional
            Help to display if the environment variable is not set. (the default is None)
        description : str, optional
            Prompt to display with the text box.
            (the default is "Enter the value: ")
            "prompt" is a alias for this parameter
        auto_display : bool, optional
            Whether to display on instantiation (the default is False)

        """
        env_val = os.environ.get(env_var)
        self._name = env_var
        self._value = ""
        description = kwargs.pop("prompt", description)

        # Call superclass to register
        super().__init__(id_vals=[env_var, description], val_attrs=["_value"], **kwargs)

        # Use the registed widget "remembered" value but if the environment
        # variable is set override with this value.
        if env_val is not None:
            self._value = env_val

        if not self._value and help_str is not None:
            display(widgets.HTML(value=help_str))

        self._w_text = widgets.Text(
            value=self._value,
            description=description,
            layout=Layout(width="50%"),
            style={"description_width": "initial"},
        )

        self._w_save_button = widgets.Button(
            description="Set",
            layout=Layout(width="10%", disabled=False),
            style={"description_width": "initial"},
        )
        self._w_check_save = widgets.Checkbox(
            value=True, description="Save as environment var", disabled=False
        )
        self._w_save_button.on_click(self._on_save_button_clicked)
        self._hbox = widgets.HBox(
            [self._w_text, self._w_save_button, self._w_check_save]
        )

        if auto_display:
            self.display()

    @property
    def value(self):
        """Get the current value of the key."""
        self._value = self._w_text.value or ""
        return self._value.strip()

    @property
    def name(self):
        """Get the current name of the key."""
        return self._name

    @property
    def layout(self):
        """Return underlying widget collection."""
        return self._hbox

    def _on_save_button_clicked(self, button):
        del button
        self._value = self.value
        if self._w_check_save.value:
            os.environ[self._name] = self.value



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Mixin Classes for Sentinel Incident Features."""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Callable
from uuid import UUID, uuid4

import httpx
import pandas as pd
from azure.common.exceptions import CloudError
from IPython.display import display
from typing_extensions import Self

from msticpy.context.azure.sentinel_bookmarks import SentinelBookmarksMixin

from ..._version import VERSION
from ...common.exceptions import MsticpyUserError
from .azure_data import get_api_headers
from .sentinel_utils import (
    _azs_api_result_to_df,
    extract_sentinel_response,
    get_http_timeout,
)

if TYPE_CHECKING:
    from datetime import datetime

__version__ = VERSION
__author__ = "Pete Bryan"

logger: logging.Logger = logging.getLogger(__name__)


class SentinelIncidentsMixin(SentinelBookmarksMixin):
    """Mixin class for Sentinel Incidents feature integrations."""

    def get_incident(  # noqa:PLR0913
        self: Self,
        incident: str,
        *,
        entities: bool = False,
        alerts: bool = False,
        comments: bool = False,
        bookmarks: bool = False,
    ) -> pd.DataFrame:
        """
        Get details on a specific incident.

        Parameters
        ----------
        incident : str
            Incident ID GUID.
        entities : bool, optional
            If True include all entities in the response. Default is False.
        alerts : bool, optional
            If True include all alerts in the response. Default is False.
        comments: bool, optional
             If True include all comments in the response. Default is False.
        bookmarks: bool, optional
             If True include all bookmarks in the response. Default is False.

        Returns
        -------
        pd.DataFrame
            Table containing incident details.

        Raises
        ------
        CloudError
            If incident could not be retrieved.

        """
        incident_id: str = self._get_incident_id(incident)
        incident_url: str = self.sent_urls["incidents"] + f"/{incident_id}"
        response: httpx.Response = super(SentinelBookmarksMixin, self)._get_items(
            incident_url,
        )
        if not response.is_success:
            raise CloudError(response=response)

        incident_df: pd.DataFrame = _azs_api_result_to_df(response)

        if entities:
            incident_df["Entities"] = [self.get_entities(incident_id)]

        if alerts:
            incident_df["Alerts"] = [self.get_incident_alerts(incident_id)]

        if comments:
            incident_df["Bookmarks"] = [self.get_incident_comments(incident_id)]

        if bookmarks:
            incident_df["Bookmarks"] = [self.get_incident_bookmarks(incident_id)]

        return incident_df

    def get_entities(self: Self, incident: str) -> list:
        """
        Get the entities from an incident.

        Parameters
        ----------
        incident : str
            Incident GUID or Name .

        Returns
        -------
        list
            A list of entities.

        """
        self.check_connected()
        incident_id: str = self._get_incident_id(incident)
        entities_url: str = self.sent_urls["incidents"] + f"/{incident_id}/entities"
        ent_parameters: dict[str, str] = {"api-version": "2021-04-01"}
        if not self._token:
            err_msg = "Token not found, can't get entities."
            raise ValueError(err_msg)
        ents: httpx.Response = httpx.post(
            entities_url,
            headers=get_api_headers(self._token),
            params=ent_parameters,
            timeout=get_http_timeout(),
        )
        return (
            [(ent["kind"], ent["properties"]) for ent in ents.json()["entities"]]
            if ents.is_success
            else []
        )

    def get_incident_alerts(self: Self, incident: str) -> list:
        """
        Get the alerts from an incident.

        Parameters
        ----------
        incident : str
            Incident GUID or Name.

        Returns
        -------
        list
            A list of alerts.

        """
        self.check_connected()
        incident_id: str = self._get_incident_id(incident)
        alerts_url: str = self.sent_urls["incidents"] + f"/{incident_id}/alerts"
        alerts_parameters: dict[str, str] = {"api-version": "2021-04-01"}
        if not self._token:
            err_msg = "Token not found, can't get incident alerts."
            raise ValueError(err_msg)
        alerts_resp: httpx.Response = httpx.post(
            alerts_url,
            headers=get_api_headers(self._token),
            params=alerts_parameters,
            timeout=get_http_timeout(),
        )
        return (
            [
                {
                    "ID": alert["properties"]["systemAlertId"],
                    "Name": alert["properties"]["alertDisplayName"],
                    "AlertProperties": alert["properties"],
                }
                for alert in alerts_resp.json()["value"]
            ]
            if alerts_resp.is_success
            else []
        )

    def get_incident_comments(self: Self, incident: str) -> list:
        """
        Get the comments from an incident.

        Parameters
        ----------
        incident : str
            Incident GUID or Name.

        Returns
        -------
        list
            A list of comments.

        """
        incident_id: str = self._get_incident_id(incident)
        comments_url: str = self.sent_urls["incidents"] + f"/{incident_id}/comments"
        comments_response: httpx.Response = self._get_items(
            comments_url,
            {"api-version": "2020-04-01"},
        )
        comment_details: dict[str, Any] = comments_response.json()
        return (
            [
                {
                    "Message": comment["properties"]["message"],
                    "Author": comment["properties"]["author"]["name"],
                }
                for comment in comment_details["value"]
            ]
            if comments_response.is_success
            else []
        )

    def get_incident_bookmarks(self: Self, incident: str) -> list:
        """
        Get the comments from an incident.

        Parameters
        ----------
        incident : str
            Incident GUID or name.

        Returns
        -------
        list
            A list of bookmarks.

        """
        bookmarks_list: list[dict[str, Any]] = []
        incident_id: str = self._get_incident_id(incident)
        relations_url: str = self.sent_urls["incidents"] + f"/{incident_id}/relations"
        relations_response: httpx.Response = self._get_items(
            relations_url,
            {"api-version": "2020-04-01"},
        )
        if relations_response.is_success and relations_response.json()["value"]:
            for relationship in relations_response.json()["value"]:
                if (
                    relationship["properties"]["relatedResourceType"]
                    == "Microsoft.SecurityInsights/Bookmarks"
                ):
                    bkmark_id: str = relationship["properties"]["relatedResourceName"]
                    bookmarks_df: pd.DataFrame = self.list_bookmarks()
                    bookmark: pd.Series = bookmarks_df[
                        bookmarks_df["name"] == bkmark_id
                    ].iloc[0]
                    bookmarks_list.append(
                        {
                            "Bookmark ID": bkmark_id,
                            "Bookmark Title": bookmark["properties.displayName"],
                        },
                    )

        return bookmarks_list

    def update_incident(
        self: Self,
        incident_id: str,
        update_items: dict,
    ) -> str:
        """
        Update properties of an incident.

        Parameters
        ----------
        incident_id : str
            Incident ID GUID.
        update_items : dict
            Dictionary of properties to update and their values.
            https://docs.microsoft.com/rest/api/securityinsights/
            stable/incidents/create-or-update

        Raises
        ------
        CloudError
            If incident could not be updated.

        """
        self.check_connected()
        incident_dets: pd.DataFrame = self.get_incident(incident_id)
        incident_url: str = self.sent_urls["incidents"] + f"/{incident_id}"
        params: dict[str, str] = {"api-version": "2020-01-01"}
        if "title" not in update_items:
            update_items["title"] = incident_dets.iloc[0]["properties.title"]
        if "status" not in update_items:
            update_items["status"] = incident_dets.iloc[0]["properties.status"]
        data: dict[str, Any] = extract_sentinel_response(
            update_items,
            props=True,
            etag=incident_dets.iloc[0]["etag"],
        )
        if not self._token:
            err_msg = "Token not found, can't update incident."
            raise ValueError(err_msg)
        response: httpx.Response = httpx.put(
            incident_url,
            headers=get_api_headers(self._token),
            params=params,
            content=str(data),
            timeout=get_http_timeout(),
        )
        if response.status_code not in (200, 201):
            raise CloudError(response=response)
        logger.info("Incident updated.")
        return response.json().get("name")

    def create_incident(  # pylint: disable=too-many-arguments, too-many-locals #noqa:PLR0913
        self: Self,
        title: str,
        severity: str,
        status: str = "New",
        description: str | None = None,
        first_activity_time: datetime | None = None,
        last_activity_time: datetime | None = None,
        labels: list[dict[str, Any]] | None = None,
        bookmarks: list[str] | None = None,
    ) -> str | None:
        """
        Create a Sentinel Incident.

        Parameters
        ----------
        title : str
            The title of the incident to create
        severity : str
            The severity to assign the incident, options are:
               Informational, Low, Medium, High
        status : str, optional
            The status to assign the incident, by default "New"
            Options are: New, Active, Closed
        description : str, optional
            A description of the incident, by default None
        first_activity_time : datetime, optional
            The start time of the incident activity, by default None
        last_activity_time : datetime, optional
            The end time of the incident activity, by default None
        labels : List, optional
            Any labels to apply to the incident, by default None
        bookmarks : List, optional
            A list of bookmark GUIDS you want to associate with the incident

        Returns
        -------
        Optional[str]
            The name/ID of the incident.

        Raises
        ------
        CloudError
            If the API returns an error

        """
        self.check_connected()
        incident_id: UUID = uuid4()
        incident_url: str = self.sent_urls["incidents"] + f"/{incident_id}"
        params: dict[str, str] = {"api-version": "2020-01-01"}
        data_items: dict[str, str | list] = {
            "title": title,
            "severity": severity.capitalize(),
            "status": status.capitalize(),
        }
        if description:
            data_items["description"] = description
        if labels:
            labels = [{"labelName": lab, "labelType": "User"} for lab in labels]
            data_items["labels"] = labels
        if first_activity_time:
            data_items["firstActivityTimeUtc"] = first_activity_time.isoformat()
        if last_activity_time:
            data_items["lastActivityTimeUtc"] = last_activity_time.isoformat()
        data: dict[str, Any] = extract_sentinel_response(data_items, props=True)
        if not self._token:
            err_msg: str = "Token not found, can't create incident."
            raise ValueError(err_msg)
        response: httpx.Response = httpx.put(
            incident_url,
            headers=get_api_headers(self._token),
            params=params,
            content=str(data),
            timeout=get_http_timeout(),
        )
        if not response.is_success:
            raise CloudError(response=response)
        if bookmarks:
            for mark in bookmarks:
                relation_id: UUID = uuid4()
                bookmark_id: str = self._get_bookmark_id(mark)
                mark_res_id: str = self.sent_urls["bookmarks"] + f"/{bookmark_id}"
                relations_url: str = incident_url + f"/relations/{relation_id}"
                bkmark_data_items: dict[str, Any] = {"relatedResourceId": mark_res_id}
                data = extract_sentinel_response(bkmark_data_items, props=True)
                params = {"api-version": "2021-04-01"}
                if not self._token:
                    err_msg = "Token not found, can't create relations."
                    raise ValueError(err_msg)
                response = httpx.put(
                    relations_url,
                    headers=get_api_headers(self._token),
                    params=params,
                    content=str(data),
                    timeout=get_http_timeout(),
                )
        logger.info("Incident created.")
        return response.json().get("name")

    def _get_incident_id(self: Self, incident: str) -> str:
        """
        Get an incident ID.

        Parameters
        ----------
        incident : str
            An incident identifier

        Returns
        -------
        str
            The Incident GUID

        Raises
        ------
        MsticpyUserError
            If incident can't be found or multiple matching incidents found.

        """
        try:
            UUID(incident)
        except ValueError as incident_name:
            incidents: pd.DataFrame = self.list_incidents()
            filtered_incidents: pd.DataFrame = incidents[
                incidents["properties.title"].str.contains(incident)
            ]
            if len(filtered_incidents) > 1:
                display(filtered_incidents[["name", "properties.title"]])
                err_msg: str = "More than one incident found, please specify by GUID"
                raise MsticpyUserError(err_msg) from incident_name
            if (
                not isinstance(filtered_incidents, pd.DataFrame)
                or filtered_incidents.empty
            ):
                err_msg = f"Incident {incident} not found"
                raise MsticpyUserError(err_msg) from incident_name
            return filtered_incidents["name"].iloc[0]
        return incident

    def post_comment(
        self: Self,
        incident_id: str,
        comment: str,
    ) -> str:
        """
        Write a comment for an incident.

        Parameters
        ----------
        incident_id : str
            Incident ID GUID.
        comment : str
            Comment message to post.

        Raises
        ------
        CloudError
            If message could not be posted.

        """
        self.check_connected()
        comment_url: str = (
            self.sent_urls["incidents"] + f"/{incident_id}/comments/{uuid4()}"
        )
        params: dict[str, str] = {"api-version": "2020-01-01"}
        data: dict[str, Any] = extract_sentinel_response({"message": comment})
        if not self._token:
            err_msg = "Token not found, can't post comment."
            raise ValueError(err_msg)
        response: httpx.Response = httpx.put(
            comment_url,
            headers=get_api_headers(self._token),
            params=params,
            content=str(data),
            timeout=get_http_timeout(),
        )
        if not response.is_success:
            raise CloudError(response=response)
        logger.info("Comment posted.")
        return response.json().get("name")

    def add_bookmark_to_incident(self: Self, incident: str, bookmark: str) -> str:
        """
        Add a bookmark to an incident.

        Parameters
        ----------
        incident : str
            Either an incident name or an incident GUID
        bookmark : str
            Either a bookmark name or bookmark GUID

        Raises
        ------
        CloudError
            If API returns error

        """
        self.check_connected()
        incident_id: str = self._get_incident_id(incident)
        incident_url: str = self.sent_urls["incidents"] + f"/{incident_id}"
        bookmark_id: str = self._get_bookmark_id(bookmark)
        mark_res_id: str = self.sent_urls["bookmarks"] + f"/{bookmark_id}"
        relations_id: UUID = uuid4()
        bookmark_url: str = incident_url + f"/relations/{relations_id}"
        bkmark_data_items: dict[str, Any] = {
            "relatedResourceId": mark_res_id.split(self.base_url)[1],
        }
        data: dict[str, Any] = extract_sentinel_response(bkmark_data_items, props=True)
        params: dict[str, str] = {"api-version": "2021-04-01"}
        if not self._token:
            err_msg = "Token not found, can't add bookmark to incident."
            raise ValueError(err_msg)
        response: httpx.Response = httpx.put(
            bookmark_url,
            headers=get_api_headers(self._token),
            params=params,
            content=str(data),
            timeout=get_http_timeout(),
        )
        if not response.is_success:
            raise CloudError(response=response)
        logger.info("Bookmark added to incident.")
        return response.json().get("name")

    def list_incidents(self: Self, params: dict | None = None) -> pd.DataFrame:
        """
        Get a list of incident for a Sentinel workspace.

        Parameters
        ----------
        params : Optional[dict], optional
            Additional parameters to pass to the API call, by default None

        Returns
        -------
        pd.DataFrame
            A table of incidents.

        Raises
        ------
        CloudError
            If incidents could not be retrieved.

        """
        if params is None:
            params = {"$top": 50}
        return self._list_items(item_type="incidents", params=params)

    get_incidents: Callable[..., pd.DataFrame] = list_incidents



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
AbuseIPDB Provider.

Input can be a single IoC observable or a pandas DataFrame containing
multiple observables. Processing may require an API key and
processing performance may be limited to a specific number of
requests per minute for the account type that you have.

"""
from __future__ import annotations

from typing import Any, ClassVar

from typing_extensions import Self

from ..._version import VERSION
from ...common.utility import export
from ..http_provider import APILookupParams
from .result_severity import ResultSeverity
from .ti_http_provider import HttpTIProvider

__version__ = VERSION
__author__ = "Rubén Revuelta Briz"


@export
class AbuseIPDB(HttpTIProvider):
    """AbuseIPDB Lookup."""

    PROVIDER_NAME: ClassVar[str] = "AbuseIPDB"

    _BASE_URL: ClassVar[str] = "https://api.abuseipdb.com/"

    _QUERIES: ClassVar[dict[str, APILookupParams]] = {
        "ipv4": APILookupParams(
            path="api/v2/check",
            params={"ipAddress": "{observable}"},
            headers={"Key": "{AuthKey}", "Accept": "application/json"},
        ),
    }

    HIGH_SEVERITY: ClassVar[int] = 50

    # aliases
    _QUERIES["ipv6"] = _QUERIES["ipv4"]

    def parse_results(self: Self, response: dict) -> tuple[bool, ResultSeverity, Any]:
        """Return the details of the response."""
        if self._failed_response(response) or not isinstance(
            response["RawResult"],
            dict,
        ):
            return False, ResultSeverity.information, "Not found."

        data: dict[str, Any] = response["RawResult"]["data"]
        result_dict: dict[str, Any] = {
            "countryCode": data.get("countryCode"),
            "usage": data.get("usageType"),
            "isp": data.get("isp"),
            "domain": data.get("domain"),
            "hostNames": data.get("hostnames"),
            "lastReportedAt": data.get("lastReportedAt"),
        }

        score: int = data.get("abuseConfidenceScore", 0)

        if score > self.HIGH_SEVERITY:
            result_severity = ResultSeverity.high
        elif score > 0:
            result_severity = ResultSeverity.warning
        else:
            result_severity = ResultSeverity.information

        return (True, result_severity, result_dict)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# Author: Thomas Roccia - @fr0gger_
# --------------------------------------------------------------------------
"""Pulsedive TI Provider."""
from __future__ import annotations

from enum import Enum
from time import sleep
from typing import TYPE_CHECKING, Any, ClassVar, NamedTuple

import httpx
import pandas as pd
from pandas import json_normalize
from typing_extensions import Self

from ..._version import VERSION
from ...common.pkg_config import get_http_timeout
from ...common.provider_settings import get_provider_settings
from ...common.utility import mp_ua_header
from ..http_provider import APILookupParams
from .result_severity import LookupResult, ResultSeverity
from .ti_http_provider import HttpTIProvider

if TYPE_CHECKING:
    from ...common.provider_settings import ProviderSettings

__version__ = VERSION
__author__ = "Thomas Roccia | @fr0gger_"

_BASE_URL = "https://pulsedive.com/api/"

_QUERY_OBJECTS_MAPPINGS: dict[str, dict[str, str]] = {
    "indicator": {"indicator": "observable"},
    "threat": {"threat": "observable"},
    "explore": {"q": "query"},
    "scan": {"scan": "observable"},
}


class PDQuery(NamedTuple):
    """Query object for Pulsedive queries."""

    query_type: str
    data: str


class PDEntityType(Enum):
    """
    Enum class for Pulsedive entity types.

    This class defines the different types of entities that can be queried through
    the Pulsedive API. The available options are:

    INDICATOR: Queries information about a specific indicator of compromise (IOC)
    THREAT: Queries information about a specific threat
    EXPLORE: Executes a general search across the Pulsedive data set
    SCAN: Submits an indicator of compromise (IOC) for scanning

    """

    INDICATOR = "indicator"
    THREAT = "threat"
    EXPLORE = "explore"
    SCAN = "scan"


class PDlookup:
    """
    PDlookup: A class to interact with the Pulsedive API.

    This class allows you to look up indicators of compromise (IOCs) and
    explore different types of data available in the Pulsedive API.

    """

    _SUPPORTED_PD_TYPES: ClassVar[set[PDEntityType]] = {
        PDEntityType.INDICATOR,
        PDEntityType.THREAT,
        PDEntityType.EXPLORE,
        PDEntityType.SCAN,
    }

    def __init__(self: PDlookup, pd_key: str | None = None) -> None:
        """
        Init function to get the API key if necessary.

        Parameters
        ----------
        pd_key: str
            An API key for the Pulsedive API.

        """
        self.pd_key: str | None = pd_key or self._get_pd_api_key()

    def lookup_ioc(
        self: Self,
        observable: str,
        pd_type: str = "indicator",
    ) -> pd.DataFrame:
        """
        Lookup an indicator of compromise (IOC) in the Pulsedive API.

        Parameters
        ----------
        observable : str
            The IOC to lookup.
        pd_type : str
            The lookup type to perform, default is "indicator".

        Returns
        -------
        pd.DataFrame
            A Pandas DataFrame with the API's response.

        Notes
        -----
        In the case of `pd_type="explore"`, the `observable`
        value should be a pulsedive query string.
        See https://pulsedive.com/explore/

        See Also
        --------
        PDlookup.lookup_threat
        PDlookup.explore
        PDlookup.scan

        """
        if pd_type == PDEntityType.THREAT.value:
            return self.lookup_threat(observable=observable)
        if pd_type == PDEntityType.EXPLORE.value:
            return self.explore(query=observable)
        if pd_type == PDEntityType.SCAN.value:
            return self.scan(observable=observable)
        return self._make_pd_request(
            _build_query_string(observable, pd_type=PDEntityType.INDICATOR.value),
        )

    def lookup_threat(self: Self, observable: str) -> pd.DataFrame:
        """
        Lookup a Threat name in the Pulsedive API.

        Parameters
        ----------
        observable : str
            The IOC to lookup.

        Returns
        -------
        pd.DataFrame
            A Pandas DataFrame with the API's response.

        """
        return self._make_pd_request(
            _build_query_string(observable, pd_type=PDEntityType.THREAT.value),
        )

    def explore(self: Self, query: str) -> pd.DataFrame:
        """
        Perform a search query on the Pulsedive API.

        Parameters
        ----------
        query : str
            The query to perform.

        Returns
        -------
        pd.DataFrame
            A Pandas DataFrame with the API's response.

        """
        return self._make_pd_request(
            _build_query_string(query, pd_type=PDEntityType.EXPLORE.value),
        )

    def scan(self: Self, observable: str) -> pd.DataFrame:
        """
        Scan an observable in the Pulsedive API.

        Parameters
        ----------
        observable : str
            The observable to scan.

        Returns
        -------
        pd.DataFrame
            A Pandas DataFrame with the API's response.

        """
        return self._make_pd_request(
            _build_query_string(observable, pd_type=PDEntityType.SCAN.value),
        )

    def _check_valid_type(self: Self, pd_type: str) -> None:
        if pd_type not in _QUERY_OBJECTS_MAPPINGS:
            err_msg: str = (
                f"Invalid pd_type: {pd_type}."
                f"Should be one of: {', '.join(_QUERY_OBJECTS_MAPPINGS.keys())}"
            )
            raise ValueError(err_msg)

    def _make_pd_request(self: Self, pd_query: PDQuery) -> pd.DataFrame:
        """
        Make a request to the PulseDive API with the provided data.

        Parameters
        ----------
        pd_query : PDQuery
            Request data

        Returns
        -------
        pd.DataFrame
            Dataframe containing the results of the API call

        """
        if pd_query.query_type == "scan":
            # if the key is "scan", make a POST request to the analyze endpoint
            data: dict[str, Any] = {
                "value": pd_query.data,
                "probe": 1,
                **self._get_default_params(),
            }
            headers: dict[str, str] = {
                "Content-Type": "application/x-www-form-urlencoded",
                **mp_ua_header(),
            }
            resp: httpx.Response = httpx.post(
                f"{_BASE_URL}analyze.php",
                data=data,
                headers=headers,
            )
            if not resp.is_success:
                err_msg: str = (
                    f"API request failed with status code {resp.status_code}"
                    f" and message {resp.text}"
                )
                raise ValueError(err_msg)
            return self._poll_for_results(resp)

        if pd_query.query_type == "q":
            # if the key is "q", make a GET request to the explore endpoint
            query_url: str = f"{_BASE_URL}explore.php"
        else:
            query_url = f"{_BASE_URL}info.php"
        params: dict[str, Any] = {
            **self._get_default_params(),
            pd_query.query_type: pd_query.data,
        }
        resp = httpx.get(
            query_url,
            params=params,
            headers=mp_ua_header(),
            timeout=get_http_timeout(),
        )
        if not resp.is_success:
            err_msg = (
                f"API request failed with status code {resp.status_code}"
                f" and message {resp.text}"
            )
            raise ValueError(err_msg)
        json_resp: dict[str, Any] = resp.json()
        if pd_query.query_type == "q":
            json_resp = json_resp["results"]
        return json_normalize(json_resp)

    def _get_default_params(self: Self) -> dict[str, Any]:
        """Return default query parameters."""
        return {"pretty": 1, "key": self.pd_key}

    def _poll_for_results(self: Self, resp: httpx.Response) -> pd.DataFrame:
        """Poll API for results for this query ID."""
        # if the initial request is successful, get the qid and
        # check the status until it is 'success'
        qid = resp.json()["qid"]
        params: dict[str, Any] = {"qid": qid, **self._get_default_params()}
        headers: dict[str, str] = mp_ua_header()
        timeout: httpx.Timeout = get_http_timeout()
        url: str = f"{_BASE_URL}analyze.php"
        resp = httpx.get(
            url,
            params=params,
            headers=headers,
            timeout=timeout,
        )
        status: str = resp.json()["status"]
        while status == "processing":
            sleep(0.5)  # wait a little to avoid hammering UI
            resp = httpx.get(url, params=params, headers=headers, timeout=timeout)
            status = resp.json()["status"]

        return json_normalize(resp.json()["data"])

    def _get_pd_api_key(self: Self) -> str | None:
        """Return Pulsedive api key from configuration."""
        pd_settings: ProviderSettings | None = get_provider_settings("TIProviders").get(
            "Pulsedive",
        )
        if pd_settings and "AuthKey" in pd_settings.args:
            return pd_settings.args["AuthKey"]
        return None


def _build_query_string(data: str, pd_type: str) -> PDQuery:
    """
    Build query string for the API request based on the provided data and pd_type.

    Parameters
    ----------
    data : Any
        Data to be mapped to query_item
    pd_type : str
        type of data to map

    Returns
    -------
    Dict[str, Any]
        Query items in dictionary

    """
    query_items = _QUERY_OBJECTS_MAPPINGS[pd_type]

    # check if the provided data is empty
    if not data:
        err_msg: str = "data is empty."
        raise ValueError(err_msg)

    return PDQuery(next(iter(query_items.keys())), data)


# -------------------------------
# MSTICPY Standard TI Provider

_QUERY_DEF = APILookupParams(
    path="info.php",
    params={"indicator": "{observable}", "pretty": 1, "key": "{API_KEY}"},
)


class Pulsedive(HttpTIProvider):
    """Pulsedive TI Lookup."""

    _BASE_URL = _BASE_URL

    _QUERIES: ClassVar[dict[str, APILookupParams]] = {
        ioc_type: _QUERY_DEF for ioc_type in ("ipv4", "ipv6", "dns", "hostname", "url")
    }

    _REQUIRED_PARAMS: ClassVar[list[str]] = ["API_KEY"]
    _RISK_MAP: ClassVar[dict[str, ResultSeverity]] = {
        "high": ResultSeverity.high,
        "medium": ResultSeverity.warning,
        "low": ResultSeverity.information,
        "none": ResultSeverity.unknown,
    }

    def __init__(
        self: Pulsedive,
        *,
        timeout: int | None = None,
        ApiID: str | None = None,  # noqa: N803
        AuthKey: str | None = None,  # noqa: N803
        Instance: str | None = None,  # noqa: N803
    ) -> None:
        """Set OTX specific settings."""
        super().__init__(
            timeout=timeout,
            ApiID=ApiID,
            AuthKey=AuthKey,
            Instance=Instance,
        )
        self.require_url_encoding = True

    def parse_results(self: Self, response: dict) -> LookupResult:
        """
        Return the details of the response.

        Parameters
        ----------
        response : Dict
            The returned data response

        Returns
        -------
        Tuple[bool, ResultSeverity, Any]
            bool = positive or negative hit
            ResultSeverity = enumeration of severity
            Object with match details

        """
        if self._failed_response(response) or not isinstance(
            response["RawResult"],
            dict,
        ):
            return LookupResult(
                status=False,
                severity=ResultSeverity.information,
                details={"status": "Not found."},
            )
        severity: ResultSeverity = self._RISK_MAP[response.get("risk", "none")]

        details: dict[str, Any] = {
            "indicator_id": response.get("iid"),
            "threats": response.get("threats", []),
            "last_seen": response.get("stamp_seen", ""),
            "risk_factors": response.get("riskfactors", {}),
            "comments": response.get("comments", {}),
        }

        return LookupResult(status=True, severity=severity, details=details)



"""Cloud mappings for offline use."""

# pylint: disable=line-too-long

cloud_mappings_offline = {
    # All endpoints were retrieved with api-version 2023-01-01
    "global": {
        "portal": "https://portal.azure.com",
        "authentication": {
            "loginEndpoint": "https://login.microsoftonline.com",
            "audiences": [
                "https://management.core.windows.net/",
                "https://management.azure.com/",
            ],
            "tenant": "common",
            "identityProvider": "AAD",
        },
        "media": "https://rest.media.azure.net",
        "graphAudience": "https://graph.windows.net/",
        "graph": "https://graph.windows.net/",
        "name": "AzureCloud",
        "suffixes": {
            "azureDataLakeStoreFileSystem": "azuredatalakestore.net",
            "acrLoginServer": "azurecr.io",
            "sqlServerHostname": "database.windows.net",
            "azureDataLakeAnalyticsCatalogAndJob": "azuredatalakeanalytics.net",
            "keyVaultDns": "vault.azure.net",
            "storage": "core.windows.net",
            "azureFrontDoorEndpointSuffix": "azurefd.net",
            "storageSyncEndpointSuffix": "afs.azure.net",
            "mhsmDns": "managedhsm.azure.net",
            "mysqlServerEndpoint": "mysql.database.azure.com",
            "postgresqlServerEndpoint": "postgres.database.azure.com",
            "mariadbServerEndpoint": "mariadb.database.azure.com",
            "synapseAnalytics": "dev.azuresynapse.net",
            "attestationEndpoint": "attest.azure.net",
        },
        "batch": "https://batch.core.windows.net/",
        "resourceManager": "https://management.azure.com/",
        "vmImageAliasDoc": "https://raw.githubusercontent.com/Azure/azure-rest-api-specs/master/arm-compute/quickstart-templates/aliases.json",
        "activeDirectoryDataLake": "https://datalake.azure.net/",
        "sqlManagement": "https://management.core.windows.net:8443/",
        "microsoftGraphResourceId": "https://graph.microsoft.com/",
        "appInsightsResourceId": "https://api.applicationinsights.io",
        "appInsightsTelemetryChannelResourceId": (
            "https://dc.applicationinsights.azure.com/v2/track"
        ),
        "attestationResourceId": "https://attest.azure.net",
        "synapseAnalyticsResourceId": "https://dev.azuresynapse.net",
        "logAnalyticsResourceId": "https://api.loganalytics.io",
        "ossrDbmsResourceId": "https://ossrdbms-aad.database.windows.net",
    },
    "usgov": {
        "portal": "https://portal.azure.us",
        "authentication": {
            "loginEndpoint": "https://login.microsoftonline.us",
            "audiences": [
                "https://management.core.usgovcloudapi.net",
                "https://management.usgovcloudapi.net",
            ],
            "tenant": "common",
            "identityProvider": "AAD",
        },
        "media": "https://rest.media.usgovcloudapi.net",
        "graphAudience": "https://graph.windows.net/",
        "graph": "https://graph.windows.net/",
        "name": "AzureUSGovernment",
        "suffixes": {
            "acrLoginServer": "azurecr.us",
            "sqlServerHostname": "database.usgovcloudapi.net",
            "keyVaultDns": "vault.usgovcloudapi.net",
            "storage": "core.usgovcloudapi.net",
            "storageSyncEndpointSuffix": "afs.azure.us",
            "mhsmDns": "managedhsm.usgovcloudapi.net",
            "mysqlServerEndpoint": "mysql.database.usgovcloudapi.net",
            "postgresqlServerEndpoint": "postgres.database.usgovcloudapi.net",
            "mariadbServerEndpoint": "mariadb.database.usgovcloudapi.net",
            "synapseAnalytics": "dev.azuresynapse.usgovcloudapi.net",
        },
        "batch": "https://batch.core.usgovcloudapi.net",
        "resourceManager": "https://management.usgovcloudapi.net",
        "vmImageAliasDoc": "https://raw.githubusercontent.com/Azure/azure-rest-api-specs/master/arm-compute/quickstart-templates/aliases.json",
        "sqlManagement": "https://management.core.usgovcloudapi.net:8443",
        "microsoftGraphResourceId": "https://graph.microsoft.us/",
        "appInsightsResourceId": "https://api.applicationinsights.us",
        "appInsightsTelemetryChannelResourceId": (
            "https://dc.applicationinsights.us/v2/track"
        ),
        "synapseAnalyticsResourceId": "https://dev.azuresynapse.usgovcloudapi.net",
        "logAnalyticsResourceId": "https://api.loganalytics.us",
        "ossrDbmsResourceId": "https://ossrdbms-aad.database.usgovcloudapi.net",
    },
    "cn": {
        "portal": "https://portal.azure.cn",
        "authentication": {
            "loginEndpoint": "https://login.chinacloudapi.cn",
            "audiences": [
                "https://management.core.chinacloudapi.cn",
                "https://management.chinacloudapi.cn",
            ],
            "tenant": "common",
            "identityProvider": "AAD",
        },
        "media": "https://rest.media.chinacloudapi.cn",
        "graphAudience": "https://graph.chinacloudapi.cn/",
        "graph": "https://graph.chinacloudapi.cn/",
        "name": "AzureChinaCloud",
        "suffixes": {
            "acrLoginServer": "azurecr.cn",
            "sqlServerHostname": "database.chinacloudapi.cn",
            "keyVaultDns": "vault.azure.cn",
            "storage": "core.chinacloudapi.cn",
            "mhsmDns": "managedhsm.azure.cn",
            "mysqlServerEndpoint": "mysql.database.chinacloudapi.cn",
            "postgresqlServerEndpoint": "postgres.database.chinacloudapi.cn",
            "synapseAnalytics": "dev.azuresynapse.azure.cn",
        },
        "batch": "https://batch.chinacloudapi.cn",
        "resourceManager": "https://management.chinacloudapi.cn",
        "vmImageAliasDoc": "https://raw.githubusercontent.com/Azure/azure-rest-api-specs/master/arm-compute/quickstart-templates/aliases.json",
        "sqlManagement": "https://management.core.chinacloudapi.cn:8443",
        "microsoftGraphResourceId": "https://microsoftgraph.chinacloudapi.cn",
        "appInsightsResourceId": "https://api.applicationinsights.azure.cn",
        "appInsightsTelemetryChannelResourceId": (
            "https://dc.applicationinsights.azure.cn/v2/track"
        ),
        "synapseAnalyticsResourceId": "https://dev.azuresynapse.azure.cn",
        "logAnalyticsResourceId": "https://api.loganalytics.azure.cn",
        "ossrDbmsResourceId": "https://ossrdbms-aad.database.chinacloudapi.cn",
    },
}



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Mixin Classes for Sentinel Analytics Features."""
from __future__ import annotations

import logging
from typing import Any, Callable
from uuid import UUID, uuid4

import httpx
import pandas as pd
from azure.common.exceptions import CloudError
from IPython.display import display
from typing_extensions import Self

from ..._version import VERSION
from ...common.exceptions import MsticpyUserError
from .azure_data import get_api_headers
from .sentinel_utils import (
    SentinelUtilsMixin,
    extract_sentinel_response,
    get_http_timeout,
)

__version__ = VERSION
__author__ = "Pete Bryan"

logger: logging.Logger = logging.getLogger(__name__)


class SentinelHuntingMixin(SentinelUtilsMixin):
    """Mixin class for Sentinel Hunting feature integrations."""

    def list_hunting_queries(self: Self) -> pd.DataFrame:
        """
        Return all custom hunting queries in a Microsoft Sentinel workspace.

        Returns
        -------
        pd.DataFrame
            A table of the custom hunting queries.

        """
        saved_query_df: pd.DataFrame = self._list_items(
            item_type="ss_path",
            api_version="2020-08-01",
        )
        return saved_query_df[
            saved_query_df["properties.category"] == "Hunting Queries"
        ]

    get_hunting_queries: Callable[..., pd.DataFrame] = list_hunting_queries

    def list_saved_queries(self: Self) -> pd.DataFrame:
        """
        Return all saved queries in a Microsoft Sentinel workspace.

        Returns
        -------
        pd.DataFrame
            A table of the custom hunting queries.

        """
        saved_query_df: pd.DataFrame = self._list_items(
            item_type="ss_path",
            api_version="2020-08-01",
        )
        return saved_query_df


class SentinelAnalyticsMixin(SentinelUtilsMixin):
    """Mixin class for Sentinel Analytics feature integrations."""

    def list_alert_rules(self: Self) -> pd.DataFrame:
        """
        Return all Microsoft Sentinel alert rules for a workspace.

        Returns
        -------
        pd.DataFrame
            A table of the workspace's alert rules.

        """
        return self._list_items(
            item_type="alert_rules",
            api_version="2024-01-01-preview",
        )

    def _get_template_id(
        self: Self,
        template: str,
    ) -> str:
        """
        Get an analytic template ID.

        Parameters
        ----------
        template : str
            Template ID or Name
        res_id : str
            Sentinel workspace to get template from

        Returns
        -------
        str
            Template ID

        Raises
        ------
        MsticpyUserError
            If template not found or multiple templates found.\

        """
        try:
            UUID(template)
        except ValueError as template_name:
            templates: pd.DataFrame = self.list_analytic_templates()
            template_details: pd.DataFrame = templates[
                templates["properties.displayName"].str.contains(template)
            ]
            if len(template_details) > 1:
                display(template_details[["name", "properties.displayName"]])
                err_msg: str = "More than one template found, please specify by GUID"
                raise MsticpyUserError(err_msg) from template_name
            if not isinstance(template_details, pd.DataFrame) or template_details.empty:
                err_msg = f"Template {template_details} not found"
                raise MsticpyUserError(err_msg) from template_name
            return template_details["name"].iloc[0]
        return template

    def create_analytic_rule(  # pylint: disable=too-many-arguments, too-many-locals #noqa:PLR0913
        self: Self,
        template: str | None = None,
        name: str | None = None,
        *,
        enabled: bool = True,
        query: str | None = None,
        query_frequency: str = "PT5H",
        query_period: str = "PT5H",
        severity: str = "Medium",
        suppression_duration: str = "PT1H",
        suppression_enabled: bool = False,
        trigger_operator: str = "GreaterThan",
        trigger_threshold: int = 0,
        description: str | None = None,
        tactics: list[str] | None = None,
    ) -> str | None:
        """
        Create a Sentinel Analytics Rule.

        Parameters
        ----------
        template : str, optional
            The GUID or name of a templated to create the analytic from, by default None
        name : str, optional
            The name to give the analytic, by default None
        enabled : bool, optional
            Whether you want the analytic to be enabled once deployed, by default True
        query : str, optional
            The KQL query string to use in the anlaytic, by default None
        query_frequency : str, optional
            How often the query should run in ISO8601 format, by default "PT5H"
        query_period : str, optional
            How far back the query should look in ISO8601 format, by default "PT5H"
        severity : str, optional
            The severity to raise incidents as, by default "Medium"
            Options are; Informational, Low, Medium, or High
        suppression_duration : str, optional
            How long to suppress duplicate alerts in ISO8601 format, by default "PT1H"
        suppression_enabled : bool, optional
            Whether you want to suppress duplicates, by default False
        trigger_operator : str, optional
            The operator for the trigger, by default "GreaterThan"
        trigger_threshold : int, optional
            The threshold of events required to create the incident, by default 0
        description : str, optional
            A description of the analytic, by default None
        tactics : list, optional
            A list of MITRE ATT&CK tactics related to the analytic, by default None

        Returns
        -------
        str|None
            The name/ID of the analytic rule.

        Raises
        ------
        MsticpyUserError
            If template provided isn't found.
        CloudError
            If the API returns an error.

        """
        self.check_connected()
        if template:
            template_id: str = self._get_template_id(template)
            templates: pd.DataFrame = self.list_analytic_templates()
            template_details: pd.Series = templates[
                templates["name"] == template_id
            ].iloc[0]
            name = template_details["properties.displayName"]
            query = template_details["properties.query"]
            query_frequency = template_details["properties.queryFrequency"]
            query_period = template_details["properties.queryPeriod"]
            severity = template_details["properties.severity"]
            trigger_operator = template_details["properties.triggerOperator"]
            trigger_threshold = template_details["properties.triggerThreshold"]
            description = template_details["properties.description"]
            tactics = (
                template_details["properties.tactics"]
                if not pd.isna(template_details["properties.tactics"])
                else []
            )

        if not tactics:
            tactics = []

        if not name:
            err_msg: str = "Please specify either a template ID or analytic details."
            raise MsticpyUserError(err_msg)

        rule_id: UUID = uuid4()
        analytic_url: str = self.sent_urls["alert_rules"] + f"/{rule_id}"
        data_items: dict[str, Any] = {
            "displayName": name,
            "query": query,
            "queryFrequency": query_frequency,
            "queryPeriod": query_period,
            "severity": severity,
            "suppressionDuration": suppression_duration,
            "suppressionEnabled": str(suppression_enabled).lower(),
            "triggerOperator": trigger_operator,
            "triggerThreshold": trigger_threshold,
            "description": description,
            "tactics": tactics,
            "enabled": str(enabled).lower(),
        }
        data: dict[str, Any] = extract_sentinel_response(data_items, props=True)
        data["kind"] = "Scheduled"
        params: dict[str, str] = {"api-version": "2020-01-01"}
        if not self._token:
            err_msg = "Token not found, can't create analytic rule."
            raise ValueError(err_msg)
        response: httpx.Response = httpx.put(
            analytic_url,
            headers=get_api_headers(self._token),
            params=params,
            content=str(data),
            timeout=get_http_timeout(),
        )
        if not response.is_success:
            raise CloudError(response=response)
        logger.info("Analytic Created.")
        return response.json().get("name")

    def _get_analytic_id(self: Self, analytic: str) -> str:
        """
        Get the GUID of an analytic rule.

        Parameters
        ----------
        analytic : str
            The GUID or name of the analytic

        Returns
        -------
        str
            The analytic GUID

        Raises
        ------
        MsticpyUserError
            If analytic not found or multiple matching analytics found

        """
        try:
            UUID(analytic)
        except ValueError as analytic_name:
            analytics: pd.DataFrame = self.list_analytic_rules()
            analytic_details: pd.DataFrame = analytics[
                analytics["properties.displayName"].str.contains(analytic)
            ]
            if len(analytic_details) > 1:
                display(analytic_details[["name", "properties.displayName"]])
                err_msg: str = "More than one analytic found, please specify by GUID"
                raise MsticpyUserError(err_msg) from analytic_name
            if not isinstance(analytic_details, pd.DataFrame) or analytic_details.empty:
                err_msg = f"Analytic {analytic_details} not found"
                raise MsticpyUserError(err_msg) from analytic_name
            return analytic_details["name"].iloc[0]
        return analytic

    def delete_analytic_rule(
        self: Self,
        analytic_rule: str,
    ) -> None:
        """
        Delete a deployed Analytic rule from a Sentinel workspace.

        Parameters
        ----------
        analytic_rule : str
            The GUID or name of the analytic.

        Raises
        ------
        CloudError
            If the API returns an error.

        """
        self.check_connected()
        analytic_id: str = self._get_analytic_id(analytic_rule)
        analytic_url: str = self.sent_urls["alert_rules"] + f"/{analytic_id}"
        params: dict[str, str] = {"api-version": "2020-01-01"}
        if not self._token:
            err_msg: str = "Token not found, can't delete analytic rule."
            raise ValueError(err_msg)
        response: httpx.Response = httpx.delete(
            analytic_url,
            headers=get_api_headers(self._token),
            params=params,
            timeout=get_http_timeout(),
        )
        if response.is_error:
            raise CloudError(response=response)
        logger.info("Analytic Deleted.")

    def list_analytic_templates(self) -> pd.DataFrame:
        """
        List Analytic Templates.

        Returns
        -------
        pd.DataFrame
            A DataFrame containing the analytics templates

        Raises
        ------
        CloudError
            If a valid result is not returned.

        """
        return self._list_items(item_type="alert_template")

    get_alert_rules: Callable[..., pd.DataFrame] = list_alert_rules
    list_analytic_rules: Callable[..., pd.DataFrame] = list_alert_rules
    get_analytic_rules: Callable[..., pd.DataFrame] = list_alert_rules



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Sentinel Dynamic Summary classes."""
from __future__ import annotations

import dataclasses
import json
import logging
import uuid
from datetime import datetime
from functools import singledispatchmethod
from typing import Any, Callable, ClassVar, Hashable, Iterable

import numpy as np
import pandas as pd
from typing_extensions import Self

from ..._version import VERSION
from ...common.exceptions import MsticpyUserError

__version__ = VERSION
__author__ = "Ian Hellen"

_TACTICS: tuple[str, ...] = (
    "Reconnaissance",
    "ResourceDevelopment",
    "InitialAccess",
    "Execution",
    "Persistence",
    "PrivilegeEscalation",
    "DefenseEvasion",
    "CredentialAccess",
    "Discovery",
    "LateralMovement",
    "Collection",
    "Exfiltration",
    "CommandAndControl",
    "Impact",
)
_TACTICS_DICT: dict[str, str] = {tactic.casefold(): tactic for tactic in _TACTICS}

_CLS_TO_API_MAP: dict[str, str] = {
    "summary_id": "summaryId",
    "summary_name": "summaryName",
    "azure_tenant_id": "azureTenantId",
    "tenant_id": "tenantId",
    "summary_description": "summaryDescription",
    "relation_name": "relationName",
    "relation_id": "relationId",
    "search_key": "searchKey",
    "tactics": "tactics",
    "techniques": "techniques",
    "source_info": "sourceInfo",
    "event_time_utc": "eventTimeUTC",
    "observable_type": "observableType",
    "observable_value": "observableValue",
    "packed_content": "packedContent",
    "summary_items": "rawContent",
    "summary_item_id": "summaryItemId",
}
_API_TO_CLS_MAP: dict[str, str] = {val: key for key, val in _CLS_TO_API_MAP.items()}

logger: logging.Logger = logging.getLogger(__name__)


class FieldList:
    """Class to hold field names."""

    def __init__(self: FieldList, fieldnames: Iterable[str]) -> None:
        """Add fields to field mapping."""
        self.__dict__.update({field.upper(): field for field in fieldnames})

    def __repr__(self: Self) -> str:
        """Return list of field attributes and values."""
        field_names: str = "\n    ".join(
            f"{key}='{val}'" for key, val in vars(self).items()
        )
        return f"Fields:\n    {field_names}"


@dataclasses.dataclass
class DynamicSummaryItem:  # pylint:disable=too-many-instance-attributes
    """
    DynamicSummaryItem class.

    Parameters
    ----------
    summary_item_id: Optional[str]
        The ID of the item
    relation_name: Optional[str] = None
        The name of the summary item relation
    relation_id: Optional[str] = None
        The ID of the summary item relation
    search_key: Optional[str] = None
        Searchable key value for summary item
    tactics: Union[str, list[str], None] = None
        Relevant MITRE tactics for the summary item
    techniques: Union[str, list[str], None] = None
        Relevant MITRE techniques for the summary item
    event_time_utc: Optional[datetime] = None
        Event time for the summary item
    observable_type: Optional[str] = None
        Observable type of the summary item
    observable_value: Optional[str] = None
        Observable value of the summary item
    packed_content: Dict[str, Any]
        Dictionary of item details.

    """

    fields: ClassVar[FieldList]
    summary_item_id: str | None = None
    relation_name: str | None = None
    relation_id: str | None = None
    search_key: str | None = None
    tactics: list[str] | str = dataclasses.field(default_factory=list)
    techniques: str | list[str] = dataclasses.field(default_factory=list)
    event_time_utc: datetime | None = None
    observable_type: str | None = None
    observable_value: str | None = None
    packed_content: dict[Hashable, Any] = dataclasses.field(default_factory=dict)

    def __post_init__(self: Self) -> None:
        """Initialize item ID if was not set explicitly."""
        self.summary_item_id = self.summary_item_id or str(uuid.uuid4())
        if isinstance(self.tactics, str):
            self.tactics = [self.tactics]
        self.tactics = _match_tactics(self.tactics)
        if isinstance(self.techniques, str):
            self.techniques = [self.techniques]

    def to_api_dict(self: Self) -> dict[str, Any]:
        """Return attributes as a JSON-serializable dictionary."""
        return {
            _CLS_TO_API_MAP.get(name, name): _convert_data_types(value)
            for name, value in dataclasses.asdict(self).items()
            if value is not None
        }


# Add helper class attribute for field names.
DynamicSummaryItem.fields = FieldList(
    [field.name for field in dataclasses.fields(DynamicSummaryItem)],
)


class DynamicSummary:
    """Dynamic Summary class."""

    fields = FieldList(
        [
            "summary_id",
            "summary_name",
            "summary_description",
            "tenant_id",
            "relation_name",
            "relation_id",
            "search_key",
            "tactics",
            "techniques",
            "source_info",
            "summary_items",
        ],
    )

    def __init__(  # pylint:disable=too-many-arguments #noqa:PLR0913
        self: DynamicSummary,
        summary_id: str | None = None,
        summary_name: str | None = None,
        summary_description: str | None = None,
        tenant_id: str | None = None,
        azure_tenant_id: str | None = None,
        search_key: str | None = None,
        tactics: str | list[str] | None = None,
        techniques: str | list[str] | None = None,
        source_info: dict[str, Any] | None = None,
        summary_items: (
            pd.DataFrame | Iterable[DynamicSummaryItem] | list[dict[str, Any]] | None
        ) = None,
    ) -> None:
        """
        Initialize a DynamicSummary instance.

        Parameters
        ----------
        summary_id : str, optional
            The summary UUID, by default auto-generated UUID
        summary_name : str, optional
            Name of the dynamic summary instance, by default None
        summary_description : str, optional
            Summary description, by default None
        tenant_id : str, optional
            Azure tenant ID, by default None
        azure_tenant_id : str, optional
            Azure tenant ID, by default None
        search_key : str, optional
            Search key column for the summarized data, by default None
        tactics : Union[str, list[str], None], optional
            Relevant MITRE tactics, by default None
        techniques : Union[str, list[str], None], optional
            Relevant MITRE techniques, by default None
        source_info : Dict[str, Any], optional
            Summary source info dictionary, by default None
        summary_items : Union[pd, DataFrame, Iterable[DynamicSummaryItem]
            list of summary items
        list[Dict[str, Any]]], optional
            Collection of summary items, by default None

        """
        self.summary_id: str = summary_id or str(uuid.uuid4())
        self.summary_name: str | None = summary_name
        self.summary_description: str | None = summary_description
        self.tenant_id: str | None = azure_tenant_id or tenant_id

        self.search_key: str | None = search_key
        tactics = tactics or []
        self.tactics: list[str] = _match_tactics(
            [tactics] if isinstance(tactics, str) else tactics,
        )
        techniques = techniques or []
        self.techniques: list[str] = (
            [techniques] if isinstance(techniques, str) else techniques
        )
        self.summary_items: list[DynamicSummaryItem] = []
        if summary_items is not None:
            self.add_summary_items(summary_items)
        self.source_info: dict[str, Any] = (
            source_info
            if isinstance(source_info, dict)
            else {"user_source": source_info}
        )
        self.source_info["source_pkg"] = f"MSTICPy {VERSION}"

        logger.info(
            "Dynamic summary created %s",
            summary_id or f"auto({self.summary_id})",
        )

    def __repr__(self: Self) -> str:
        """Return simple representation of instance."""
        attributes: dict[str, str | Any] = {
            key: f"'{val}'" if isinstance(val, str) else val
            for key, val in vars(self).items()
            if key != "summary_items" and val not in (None, pd.NaT, "", [])
        }
        return "\n".join(
            [
                "DynamicSummary(",
                *(f"  {key}={val}" for key, val in attributes.items()),
                f"  summary_items={len(self.summary_items)}",
                ")",
            ],
        )

    @classmethod
    def from_json(
        cls: type[Self],
        data: dict[str, Any] | str,
    ) -> Self:
        """Create new DynamicSummary instance from json string or dict."""
        if isinstance(data, str):
            try:
                data = json.loads(data)
            except json.JSONDecodeError as json_err:
                err_msg: str = "JSON Error decoding dynamic summary data"
                raise MsticpyUserError(err_msg) from json_err
            return cls.from_json(data)
        properties: dict[str, Any] = data.get("properties", data)
        summary_props: dict[str, Any] = {
            _API_TO_CLS_MAP.get(name, name): value
            for name, value in properties.items()
            if name != "rawContent"
        }
        summary = cls(**summary_props)
        summary_items: list[DynamicSummaryItem] = []
        try:
            raw_content_data: str = data.get("rawContent", "[]")
            raw_content: list[dict[str, Any]] = json.loads(raw_content_data)
        except json.JSONDecodeError as json_err:
            err_msg = "JSON Error decoding dynamic summary item data"
            raise MsticpyUserError(err_msg) from json_err
        for raw_item in raw_content:
            summary_item_props: dict[str, Any] = {
                _API_TO_CLS_MAP.get(name, name): (
                    pd.to_datetime(value) if name == "eventTimeUTC" else value
                )
                for name, value in raw_item.items()
            }
            summary_items.append(DynamicSummaryItem(**summary_item_props))
        summary.add_summary_items(summary_items)
        return summary

    @classmethod
    def new_dynamic_summary(  # pylint:disable=too-many-arguments # noqa: PLR0913
        cls: type[Self],
        summary_id: str | None = None,
        summary_name: str | None = None,
        summary_description: str | None = None,
        tenant_id: str | None = None,
        azure_tenant_id: str | None = None,
        search_key: str | None = None,
        tactics: str | list[str] | None = None,
        techniques: str | list[str] | None = None,
        source_info: dict[str, Any] | None = None,
        summary_items: (
            pd.DataFrame | Iterable[DynamicSummaryItem] | list[dict[str, Any]] | None
        ) = None,
    ) -> Self:
        """
        Return a new DynamicSummary object.

        Notes
        -----
        See the DynamicSummary class documentation for details
        of expected parameters.

        See Also
        --------
        DynamicSummary

        """
        return cls(
            summary_id=summary_id,
            summary_name=summary_name,
            summary_description=summary_description,
            tenant_id=tenant_id,
            azure_tenant_id=azure_tenant_id,
            search_key=search_key,
            tactics=tactics,
            techniques=techniques,
            source_info=source_info,
            summary_items=summary_items,
        )

    @staticmethod
    def df_to_dynamic_summaries(data: pd.DataFrame) -> list[DynamicSummary]:
        r"""
        Return a list of DynamicSummary objects from a DataFrame of summaries.

        Parameters
        ----------
        data : pd.DataFrame
            DataFrame containing dynamic summaries

        Returns
        -------
        list[DynamicSummary]
            List of Dynamic Summary objects.

        Examples
        --------
        Use the following steps to obtain a list of dynamic summaries
        from MS Sentinel and convert to DynamicSummary objects.

        .. code:: python

            query = \"\"\"
                DynamicSummary
                | where <some filter criteria>
                | where SummaryStatus == "Active" or SummaryDataType == "SummaryItem"
            \"\"\"
            data = qry_prov.exec_query(query)
            dyn_summaries = df_to_dynamic_summaries(data)

        """
        return [
            df_to_dynamic_summary(ds_data) for _, ds_data in data.groupby("SummaryId")
        ]

    @staticmethod
    def df_to_dynamic_summary(data: pd.DataFrame) -> DynamicSummary:
        r"""
        Return a single DynamicSummary object from a DataFrame.

        Parameters
        ----------
        data : pd.DataFrame
            DataFrame containing a single dynamic summary plus
            summary items.

        Returns
        -------
        DynamicSummary
            The DynamicSummary object.

        Examples
        --------
        Use the following steps to query a single dynamic summary
        from MS Sentinel and convert to a DynamicSummary object.

        .. code:: python

            query = \"\"\"
                DynamicSummary
                | where SummaryId == "26b95b5e-2645-4d33-91a7-ea3c1b8b4b8b"
                | where SummaryStatus == "Active" or SummaryDataType == "SummaryItem"
            \"\"\"
            data = qry_prov.exec_query(query)
            dyn_summaries = df_to_dynamic_summary(data)

        """
        return df_to_dynamic_summary(data)

    def add_summary_items(
        self: Self,
        data: Iterable[DynamicSummaryItem] | Iterable[dict[str, Any]] | pd.DataFrame,
        **kwargs,
    ) -> None:
        """
        Add list of DynamicSummaryItems replacing existing list.

        Parameters
        ----------
        data : Union[Iterable[DynamicSummaryItem], Iterable[Dict[str, Any]], pd.DataFrame]
            Iterable or DataFrame of DynamicSummary Items.
        summary_fields : Optional[Dict[str, str]], optional
            (only relevant if `data` is a DataFrame)
            Dictionary of mappings to extract from the DataFrame
            and use as SummaryItem properties, by default None.
            For example: {"col_a": "tactics", "col_b": "relation_name"}
            See DynamicSummaryItem for a list of available properties.

        See Also
        --------
        DynamicSummaryItem

        """
        if isinstance(data, pd.DataFrame):
            if data.empty:
                return
        elif not data:
            return
        self._add_summary_items(data, **kwargs)

    @singledispatchmethod
    def _add_summary_items(self: Self, data: list, **kwargs) -> None:
        """
        Add list of DynamicSummaryItems.

        Parameters
        ----------
        data : Union[Iterable[DynamicSummaryItem], Iterable[Dict[str, Any]]]
            Iterable of DynamicSummary Items.

        """
        del kwargs
        if isinstance(next(iter(data)), DynamicSummaryItem):
            logger.info(
                "_add_summary_items (list(DynamicSummaryItem)) items %d",
                len(data) if data else 0,
            )
            self.summary_items = list(data)
        else:
            self._add_summary_items_dict(data)

    @_add_summary_items.register(pd.DataFrame)
    def _(
        self: Self,
        data: pd.DataFrame,
        *,
        summary_fields: dict[str, str] | None = None,
        event_time_utc: str | None = None,
        search_key: str | None = None,
        **kwargs,
    ) -> None:
        """
        Add DataFrame of dynamic summary items.

        Parameters
        ----------
        data : pd.DataFrame
            DataFrame containing dynamic summary items.
        summary_fields : Optional[Dict[str, str]], optional
            dictionary of mappings to extract from the DataFrame
            and use as SummaryItem properties, by default None.
            For example: {"col_a": "tactics", "col_b": "relation_name"}
            See DynamicSummaryItem for a list of available properties.
        event_time_utc: Optional[datetime] = None
            Event time for the summary item
        search_key: Optional[str] = None
            Searchable key value for summary item

        See Also
        --------
        DynamicSummaryItem

        """
        logger.info("_add_summary_items (df) rows %d", len(data))
        for row in data.to_dict(orient="records"):
            summary_params: dict[str, Any] = {}
            if summary_fields:
                # if summary fields to map to dynamic summary item properties
                # extract these from the row dictionary first
                summary_params = {
                    field_name: row.get(column_name)
                    for field_name, column_name in summary_fields.items()
                }
            # if event time not in summary_fields, try to get from
            # kwargs or from data
            if "event_time_utc" not in summary_params:
                summary_params["event_time_utc"] = event_time_utc or row.get(
                    "TimeGenerated",
                )
            search_key_value: str | None = (
                row.get(self.search_key) if self.search_key else None
            )
            if search_key_value and not search_key:
                search_key = search_key_value
            # Create DynamicSummaryItem instance for each row
            self.summary_items.append(
                DynamicSummaryItem(
                    packed_content={
                        key: _convert_data_types(value) for key, value in row.items()
                    },
                    **{**summary_params, "search_key": search_key},
                    **kwargs,  # pass remaining kwargs as summary item properties
                ),
            )

    def _add_summary_items_dict(self: Self, data: Iterable[dict[str, Any]]) -> None:
        """
        Add DynamicSummary items from an iterable of dicts.

        Parameters
        ----------
        data : Iterable[Dict[str, Any]]
            Iterable of dictionaries containing summary item
            properties.

        """
        logger.info(
            "_add_summary_items (list(dict)) rows %d",
            len(list(data)) if data else 0,
        )
        summary_items: list[DynamicSummaryItem] = []
        for properties in data:
            # if search key specified, try to extract from packed_content field
            if (
                self.search_key
                and "search_key" not in properties
                and self.search_key in properties.get("packed_content", {})
            ):
                search_key_value: str = properties.get("packed_content", {}).get(
                    self.search_key,
                )
                if search_key_value:
                    properties["search_key"] = search_key_value
            summary_items.append(DynamicSummaryItem(**properties))
        self.summary_items = summary_items

    def append_summary_items(
        self: Self,
        data: Iterable[DynamicSummaryItem] | Iterable[dict[str, Any]] | pd.DataFrame,
        **kwargs,
    ) -> None:
        """
        Append list of DynamicSummaryItems to existing list.

        Parameters
        ----------
        data : Union[Iterable[DynamicSummaryItem], Iterable[Dict[str, Any]], pd.DataFrame]
            Iterable or DataFrame of DynamicSummary Items.
        summary_fields : Optional[Dict[str, str]], optional
            (only relevant if `data` is a DataFrame)
            Dictionary of mappings to extract from the DataFrame
            and use as SummaryItem properties, by default None.
            For example: {"col_a": "tactics", "col_b": "relation_name"}
            See DynamicSummaryItem for a list of available properties.

        See Also
        --------
        DynamicSummaryItem

        """
        current_items: list[DynamicSummaryItem] = self.summary_items
        self.add_summary_items(data, **kwargs)
        new_items: list[DynamicSummaryItem] = self.summary_items
        self.summary_items = current_items + new_items
        logger.info("append_summary_items %s", type(data))

    def to_json(self: Self) -> str:
        """Return JSON representation of DynamicSummary."""
        summary_properties: dict[str, Any] = {
            _CLS_TO_API_MAP.get(prop_name, prop_name): prop_value
            for prop_name, prop_value in self.__dict__.items()
            if prop_name in _CLS_TO_API_MAP and prop_value is not None
        }
        if self.summary_items:
            summary_properties[_CLS_TO_API_MAP["summary_items"]] = json.dumps(
                [item.to_api_dict() for item in self.summary_items],
            )
        return json.dumps(summary_properties)

    def to_json_api(self: Self) -> str:
        """Return API-ready JSON representation of DynamicSummary."""
        return f'{{"properties" : {self.to_json()} }}'

    def to_df(self: Self) -> pd.DataFrame:
        """Return summary items as DataFrame."""
        data = pd.DataFrame([item.packed_content for item in self.summary_items])
        if "TimeGenerated" in data.columns:
            return (
                data.assign(timestamp=pd.to_datetime(data["TimeGenerated"]))
                .drop(columns="TimeGenerated")
                .rename(columns={"timestamp": "TimeGenerated"})
            )
        return data


_DF_TO_CLS_MAP: dict[str, str] = {
    "TenantId": "ws_tenant_id",
    "TimeGenerated": "time_generated",
    "AzureTenantId": "tenant_id",
    "SummaryId": "summary_id",
    "SummaryItemId": "summary_item_id",
    "SummaryName": "summary_name",
    "RelationName": "relation_name",
    "RelationId": "relation_id",
    "SearchKey": "search_key",
    "CreatedBy": "created_by",
    "CreatedTimeUTC": "created_time_utc",
    "UpdatedBy": "updated_by",
    "UpdatedTimeUTC": "updated_time_utc",
    "SummaryDescription": "summary_description",
    "Tactics": "tactics",
    "Techniques": "techniques",
    "SummaryStatus": "summary_status",
    "SourceInfo": "source_info",
    "Query": "query",
    "QueryStartDate": "query_start_date",
    "QueryEndDate": "query_end_date",
    "EventTimeUTC": "event_time_utc",
    "ObservableType": "observable_type",
    "ObservableValue": "observable_value",
    "PackedContent": "packed_content",
    "SummaryDataType": "summary_data_type",
    "SourceSystem": "source_system",
    "Type": "type",
}
_CLS_TO_DF_MAP: dict[str, str] = {val: key for key, val in _DF_TO_CLS_MAP.items()}
_DF_SUMMARY_FIELDS: set[str] = {
    "TenantId",
    "TimeGenerated",
    "AzureTenantId",
    "SummaryId",
    "SummaryName",
    "RelationName",
    "RelationId",
    "SearchKey",
    "CreatedBy",
    "CreatedTimeUTC",
    "UpdatedBy",
    "UpdatedTimeUTC",
    "SummaryDescription",
    "Tactics",
    "Techniques",
    "SummaryStatus",
    "SourceInfo",
    "Query",
    "QueryStartDate",
    "QueryEndDate",
    "SummaryDataType",
}
_DF_SUMMARY_ITEM_FIELDS: set[str] = {
    "TimeGenerated",
    "SummaryItemId",
    "RelationName",
    "RelationId",
    "SearchKey",
    "CreatedBy",
    "CreatedTimeUTC",
    "UpdatedBy",
    "UpdatedTimeUTC",
    "SummaryDescription",
    "Tactics",
    "Techniques",
    "EventTimeUTC",
    "ObservableType",
    "ObservableValue",
    "PackedContent",
    "SummaryDataType",
}


def _get_summary_record(data: pd.DataFrame) -> pd.Series:
    """Return active dynamic summary header record."""
    ds_summary: pd.DataFrame = data[
        (data["SummaryDataType"] == "Summary") & (data["SummaryStatus"] == "Active")
    ]
    return ds_summary[list(_DF_SUMMARY_FIELDS)].rename(columns=_DF_TO_CLS_MAP).iloc[0]


def _get_summary_items(data: pd.DataFrame) -> pd.DataFrame:
    """Return summary item records for dynamic summary."""
    ds_summary_items: pd.DataFrame = data[data["SummaryDataType"] == "SummaryItem"]
    return ds_summary_items[list(_DF_SUMMARY_ITEM_FIELDS)].rename(
        columns=_DF_TO_CLS_MAP,
    )


def df_to_dynamic_summaries(data: pd.DataFrame) -> list[DynamicSummary]:
    r"""
    Return a list of DynamicSummary objects from a DataFrame of summaries.

    Parameters
    ----------
    data : pd.DataFrame
        DataFrame containing dynamic summaries

    Returns
    -------
    list[DynamicSummary]
        List of Dynamic Summary objects.

    Examples
    --------
    Use the following steps to obtain a list of dynamic summaries
    from MS Sentinel and convert to DynamicSummary objects.

    .. code:: python

        query = \"\"\"
            DynamicSummary
            | where <some filter criteria>
            | where SummaryStatus == "Active" or SummaryDataType == "SummaryItem"
        \"\"\"
        data = qry_prov.exec_query(query)
        dyn_summaries = df_to_dynamic_summaries(data)

    """
    return [df_to_dynamic_summary(ds_data) for _, ds_data in data.groupby("SummaryId")]


def df_to_dynamic_summary(data: pd.DataFrame) -> DynamicSummary:
    r"""
    Return a single DynamicSummary object from a DataFrame.

    Parameters
    ----------
    data : pd.DataFrame
        DataFrame containing a single dynamic summary plus
        summary items.

    Returns
    -------
    DynamicSummary
        The DynamicSummary object.

    Examples
    --------
    Use the following steps to query a single dynamic summary
    from MS Sentinel and convert to a DynamicSummary object.

    .. code:: python

        query = \"\"\"
            DynamicSummary
            | where SummaryId == "26b95b5e-2645-4d33-91a7-ea3c1b8b4b8b"
            | where SummaryStatus == "Active" or SummaryDataType == "SummaryItem"
        \"\"\"
        data = qry_prov.exec_query(query)
        dyn_summaries = df_to_dynamic_summary(data)

    """
    dyn_summary: DynamicSummary = DynamicSummary()
    dyn_summary.__dict__.update(_get_summary_record(data).to_dict())

    items_list: list[dict[Hashable, Any]] = _get_summary_items(data).to_dict(
        orient="records",
    )
    items: list[DynamicSummaryItem] = []
    for item in items_list:
        # "fields" attrib is a ClassVar
        ds_item: DynamicSummaryItem = DynamicSummaryItem()
        for key, value in item.items():
            setattr(ds_item, str(key), value)
        items.append(ds_item)
    dyn_summary.add_summary_items(items)
    return dyn_summary


def _to_datetime_utc_str(date_time: datetime | str) -> str:
    """Convert datetime to ISO date string."""
    if not isinstance(date_time, datetime):
        return date_time
    dt_str: str = date_time.isoformat()
    return dt_str.replace("+00:00", "Z") if "+00:00" in dt_str else f"{dt_str}Z"


def _convert_dict_types(input_dict: dict[Any, Any]) -> dict[Any, Any]:
    """Convert data types in dictionary members."""
    return {name: _convert_data_types(value) for name, value in input_dict.items()}


_TYPE_CONVERTER: dict[Any, Callable] = {
    np.ndarray: list,
    datetime: _to_datetime_utc_str,
    pd.Timestamp: _to_datetime_utc_str,
    dict: _convert_dict_types,
}


def _convert_data_types(
    value: str,
    type_convert: dict[type, Callable] | None = None,
) -> str:
    """Convert a type based on dictionary of converters."""
    type_convert = type_convert or {}
    type_convert.update(_TYPE_CONVERTER)
    converter: Callable | None = type_convert.get(type(value))
    return converter(value) if converter else value


def _match_tactics(tactics: Iterable[str]) -> list[str]:
    """Return case-insensitive matches for tactics list."""
    return [
        _TACTICS_DICT[tactic.casefold()]
        for tactic in tactics
        if tactic in _TACTICS_DICT
    ]



"""Context Providers sub-package."""

from __future__ import annotations

from ..._version import VERSION
from .servicenow import ServiceNow  # noqa:F401

__version__ = VERSION

CONTEXT_PROVIDERS: dict[str, tuple[str, str]] = {
    "ServiceNow": ("servicenow", "ServiceNow"),
}



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Tor Exit Nodes Provider.

Input can be a single IoC observable or a pandas DataFrame containing
multiple observables. Processing may require a an API key and
processing performance may be limited to a specific number of
requests per minute for the account type that you have.

"""
from __future__ import annotations

import contextlib
from datetime import datetime, timezone
from threading import Lock
from typing import Any, ClassVar, Iterable

import httpx
import pandas as pd
from typing_extensions import Self

from ..._version import VERSION
from ...common.pkg_config import get_http_timeout
from ...common.utility import export, mp_ua_header
from ..lookup_result import LookupStatus
from .ti_provider_base import ResultSeverity, TIProvider

__version__ = VERSION
__author__ = "Ian Hellen"


@export
class Tor(TIProvider):
    """Tor Exit Nodes Lookup."""

    _BASE_URL: ClassVar[str] = "https://check.torproject.org/exit-addresses"
    _ALT_URL: ClassVar[str] = (
        "https://raw.githubusercontent.com/SecOps-Institute/Tor-IP-Addresses"
        "/master/tor-exit-nodes.lst"
    )

    _QUERIES: ClassVar[dict[str, Any]] = {"ipv4": None, "ipv6": None}
    _nodelist: ClassVar[dict[str, dict[str, str]]] = {}
    _last_cached: datetime = datetime.min
    _cache_lock = Lock()

    @classmethod
    def _check_and_get_nodelist(cls: type[Self]) -> None:
        """Pull down Tor exit node list and save to internal attribute."""
        if cls._cache_lock.locked():
            return
        now: datetime = datetime.now(timezone.utc)
        if not cls._nodelist or (now - cls._last_cached).days > 1:
            with contextlib.suppress(ConnectionError):
                resp: httpx.Response = httpx.get(
                    cls._BASE_URL,
                    timeout=get_http_timeout(),
                    headers=mp_ua_header(),
                )
                tor_raw_list: str = resp.content.decode()
                with cls._cache_lock:
                    cls._nodelist = dict(cls._tor_splitter(tor_raw_list))
                    cls._last_cached = datetime.now(timezone.utc)
        if not cls._nodelist:
            with contextlib.suppress(ConnectionError):
                resp = httpx.get(
                    cls._ALT_URL,
                    timeout=get_http_timeout(),
                    headers=mp_ua_header(),
                )
                tor_raw_list = resp.content.decode()
                with cls._cache_lock:
                    node_dict: dict[str, Any] = {"ExitNode": True, "LastStatus": now}
                    cls._nodelist = {
                        node: node_dict for node in tor_raw_list.split("\n")
                    }
                    cls._last_cached = datetime.now(timezone.utc)

    @staticmethod
    def _tor_splitter(node_list: str) -> Iterable[tuple[str, dict[str, Any]]]:
        node_dict: dict[str, str | None] = {}
        for line in node_list.split("\n"):
            if not line:
                continue
            fields: list[str] = line.split(" ", 2)
            if fields[0] == "ExitNode":
                # new record so reset dict
                node_dict = {}
            if len(fields) > 1:
                node_dict[fields[0]] = fields[1]
            else:
                node_dict[fields[0]] = None
            if fields[0] == "ExitAddress":
                # yield tuple
                yield fields[1], node_dict

    def lookup_ioc(
        self: Self,
        ioc: str,
        ioc_type: str | None = None,
        query_type: str | None = None,
        *,
        provider_name: str | None = None,
    ) -> pd.DataFrame:
        """
        Lookup a single IoC observable.

        Parameters
        ----------
        ioc : str
            IoC Observable value
        ioc_type : str, optional
            IoC Type, by default None (type will be inferred)
        query_type : str, optional
            Specify the data subtype to be queried, by default None.
            If not specified the default record type for the IoC type
            will be returned.
        provider_name : str, optional
            Name of the provider to query

        Returns
        -------
        pd.DataFrame
            The returned results.

        """
        if not self._nodelist:
            self._check_and_get_nodelist()
        result: dict[str, Any] = self._check_ioc_type(
            ioc=ioc,
            ioc_type=ioc_type,
            query_subtype=query_type,
        )

        result["Provider"] = provider_name or self.__class__.__name__
        result["Result"] = bool(self._nodelist)
        result["Reference"] = self._BASE_URL

        if result["Status"] and not bool(self._nodelist):
            result["Status"] = LookupStatus.QUERY_FAILED.value

        if result["Status"]:
            return pd.DataFrame([result])

        tor_node = self._nodelist.get(ioc)

        if tor_node:
            result["Severity"] = ResultSeverity.warning.name
            result["Details"] = {
                "NodeID": tor_node["ExitNode"],
                "LastStatus": tor_node["LastStatus"],
            }
            result["RawResult"] = tor_node
        else:
            result["Details"] = "Not found."
        return pd.DataFrame([result])

    def parse_results(self: Self, response: dict) -> tuple[bool, ResultSeverity, Any]:
        """
        Return the details of the response.

        Parameters
        ----------
        response : Dict
            The returned data response

        Returns
        -------
        tuple[bool, ResultSeverity, Any]
            bool = positive or negative hit
            ResultSeverity = enumeration of severity
            Object with match details

        """
        del response
        return (True, ResultSeverity.information, None)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Kusto Driver subclass."""
import base64
import dataclasses
import json
import logging
from datetime import timedelta
from functools import partial
from typing import Any, Dict, List, NamedTuple, Optional, Tuple, Union

import pandas as pd
from azure.kusto.data import (
    ClientRequestProperties,
    KustoClient,
    KustoConnectionStringBuilder,
)
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.serialization import pkcs12

from ..._version import VERSION
from ...auth.azure_auth import az_connect, get_default_resource_name
from ...auth.cloud_mappings import AzureCloudConfig
from ...common.exceptions import (
    MsticpyDataQueryError,
    MsticpyMissingDependencyError,
    MsticpyNotConnectedError,
    MsticpyParameterError,
)
from ...common.provider_settings import ProviderArgs, get_protected_setting
from ...common.settings import get_config, get_http_proxies
from ...common.utility import export
from ..core.query_defns import DataEnvironment
from ..core.query_source import QuerySource
from .driver_base import DriverBase, DriverProps

# pylint: disable=ungrouped-imports
try:
    from azure.kusto.data.exceptions import KustoApiError, KustoServiceError
    from azure.kusto.data.helpers import dataframe_from_result_table
    from azure.kusto.data.response import KustoResponseDataSet
except ImportError as imp_err:
    raise MsticpyMissingDependencyError(
        "Cannot use this feature without Azure Kusto client installed",
        title="Error importing azure.kusto.data",
        packages="azure-kusto-data",
    ) from imp_err

__version__ = VERSION
__author__ = "Ian Hellen"

_HELP_URL = "https://msticpy.readthedocs.io/en/latest/DataProviders/DataProv-Kusto.html"

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class KustoConfig:
    """Kusto configuration class."""

    name: str
    cluster: str
    alias: str
    path: str
    args: ProviderArgs = dataclasses.field(default_factory=ProviderArgs)
    tenant_id: Optional[str] = None
    integrated_auth: bool = False
    cluster_groups: List[str] = dataclasses.field(default_factory=list)

    @dataclasses.dataclass
    class ConfigFields:
        """Kusto configuration fields."""

        CLUSTER = "Cluster"
        TENANT_ID = "TenantId"
        INTEG_AUTH = "IntegratedAuth"
        DEFAULTS = "ClusterDefaults"
        CLIENT_ID = "ClientId"
        CLIENT_SEC = "ClientSecret"
        ARGS = "Args"
        CLUSTER_GROUPS = "ClusterGroups"
        CERTIFICATE = "Certificate"

    # pylint: disable=no-member
    @property
    def default_db(self):
        """Return default database for this cluster."""
        return self.args.get("Database", self.args.get("DefaultDatabase"))

    # pylint: disable=unsubscriptable-object, unsupported-membership-test
    def __getattr__(self, attrib):
        """Return attribute from args if not in self."""
        if attrib in self.args:
            return self.args[attrib]
        raise AttributeError(f"Invalid attribute '{attrib}'")

    def __contains__(self, attrib):
        """Return True if attribute in self or args."""
        return attrib in self.__dict__ or attrib in self.args


@dataclasses.dataclass
class QuerySourceFields:
    """Kusto query source/yaml query fields."""

    CLUSTER = "cluster"
    CLUSTERS = "clusters"
    CLUSTER_GROUPS = "cluster_groups"
    DATA_ENVS = "data_environments"
    DATA_FAMILIES = "data_families"


class AuthParams(NamedTuple):
    """NamedTuple for auth parameters."""

    method: str
    params: Dict[str, Any]
    uri: str


KFields = KustoConfig.ConfigFields

_DEFAULT_TIMEOUT = 60 * 4
_MAX_TIMEOUT = 60 * 60


# pylint: disable=too-many-instance-attributes
@export
class AzureKustoDriver(DriverBase):
    """Kusto Driver class to execute kql queries for Azure Data Explorer."""

    def __init__(self, connection_str: Optional[str] = None, **kwargs):
        """
        Instantiate KustoDriver.

        Parameters
        ----------
        connection_str : str, optional
            Connection string

        Other Parameters
        ----------------
        debug : bool
            print out additional diagnostic information.
        timeout : int
            Query timeout in seconds, default is 240 seconds (4 minutes)
            Maximum is 3600 seconds (1 hour).
            (can be set here or in connect and overridden in query methods)
        proxies : Dict[str, str]
            Proxy settings for Kusto queries.
            Dictionary format is {protocol: proxy_url}
            Where protocol is https, http, etc. and proxy_url can contain
            optional authentication information in the format
            "https://username:password@proxy_host:port"
            If you have a proxy configuration in msticpyconfig.yaml and
            you do not want to use it, set this to an empty dictionary.
            (can be overridden in connect method)

        """
        super().__init__(**kwargs)
        if kwargs.get("debug", False):
            logger.setLevel(logging.DEBUG)
        self.environment = kwargs.get("data_environment", DataEnvironment.Kusto)
        self._strict_query_match = kwargs.get("strict_query_match", False)
        self._kusto_settings: Dict[str, Dict[str, KustoConfig]] = _get_kusto_settings()
        self._default_database: Optional[str] = None
        self._current_connection: Optional[str] = connection_str
        self._current_config: Optional[KustoConfig] = None
        self.client: Optional[KustoClient] = None
        self._az_auth_types: Optional[List[str]] = None
        self._az_tenant_id: Optional[str] = None
        self._def_timeout = min(kwargs.pop("timeout", _DEFAULT_TIMEOUT), _MAX_TIMEOUT)
        self._def_proxies = kwargs.get("proxies", get_http_proxies())

        self.add_query_filter("data_environments", "Kusto")
        self.set_driver_property(DriverProps.PUBLIC_ATTRS, self._set_public_attribs())
        self.set_driver_property(DriverProps.FILTER_ON_CONNECT, True)
        self.set_driver_property(DriverProps.EFFECTIVE_ENV, DataEnvironment.Kusto.name)
        self.set_driver_property(DriverProps.SUPPORTS_THREADING, value=True)
        self.set_driver_property(
            DriverProps.MAX_PARALLEL, value=kwargs.get("max_threads", 4)
        )
        self._loaded = True

    def _set_public_attribs(self):
        """Expose subset of attributes via query_provider."""
        return {
            "get_database_names": self.get_database_names,
            "get_database_schema": self.get_database_schema,
            "configured_clusters": self.configured_clusters,
            "cluster_uri": self.cluster_uri,
            "cluster_name": self.cluster_name,
            "cluster_config_name": self.cluster_config_name,
            "set_cluster": self.set_cluster,
            "set_database": self.set_database,
        }

    @property
    def current_connection(self) -> Optional[str]:
        """Return current connection string or URI."""
        if self._current_connection:
            return self._current_connection
        return self.cluster_uri

    @current_connection.setter
    def current_connection(self, value: str):
        """Set current connection string or URI."""
        self._current_connection = value

    @property
    def cluster_uri(self) -> str:
        """Return current cluster URI."""
        return "" if not self._current_config else self._current_config.cluster

    @property
    def cluster_name(self) -> str:
        """Return current cluster URI."""
        return self._current_config.name if self._current_config else ""

    @property
    def cluster_config_name(self) -> str:
        """Return current cluster URI."""
        if isinstance(self._current_config, KustoConfig):
            return self._current_config.alias
        return "not defined"

    @property
    def schema(self) -> Dict[str, Dict]:
        """Return schema for current database."""
        try:
            return self.get_database_schema()
        except ValueError:
            print("Default database not set - unable to retrieve schema.")
        except MsticpyNotConnectedError:
            print("Not connected to a cluster - unable to retrieve schema.")
        except MsticpyDataQueryError:
            print("Kusto Error retrieving the schema.")
        return {}

    @property
    def configured_clusters(self) -> Dict[str, KustoConfig]:
        """Return current Kusto config settings."""
        return self._kusto_settings["id"]

    def set_cluster(self, cluster: str):
        """Set the current cluster to `cluster` and connect."""
        self.connect(cluster=cluster)

    def set_database(self, database: str):
        """Set the default database to `database`."""
        self._default_database = database

    def connect(self, connection_str: Optional[str] = None, **kwargs):
        """
        Connect to data source.

        Either a connection string or a cluster name must be specified.
        The cluster name can be a short name or a full URI. If a short name,
        the cluster must be defined in the msticpyconfig.yaml file.
        In this case, the short name can be either the key of the cluster
        definition the host name part of the cluster URI.

        Parameters
        ----------
        connection_str : str
            Connect to a data source
        cluster : str, optional
            Short name or URI of cluster to connect to.

        Other Parameters
        ----------------
        database : str, optional
            Name to set the default database to.
        tenant_id : str, optional
            Azure tenant ID for the cluster.
        connection_str : str, optional
            Kusto connection string, including authentication credentials.
        auth_types: Union[str, list], optional
            Credential type or types to use for authentication.
            Use `msticpy.auth.azure_auth.list_auth_methods()` to get a list
            of available methods.
        mp_az_auth : Union[bool, str, list, None], optional
            Deprecated parameter to use MSTICPy Azure authentication.
            Values can be:
            True or "default": use the settings in msticpyconfig.yaml 'Azure' section
            str: single auth method name
            List[str]: list of acceptable auth methods
        mp_az_tenant_id: str, optional
            alias for `tenant_id`.
        timeout : int
            Query timeout in seconds, default is 240 seconds (4 minutes)
            Maximum is 3600 seconds (1 hour).
            (can be overridden in query methods)

        See Also
        --------
        msticpy.auth.azure_auth.list_auth_methods

        """
        logger.info(
            "Connecting to Kusto cluster: connection_str=%s, args=%s",
            connection_str,
            kwargs,
        )
        self._default_database = kwargs.pop("database", None)
        self._def_timeout = min(kwargs.pop("timeout", self._def_timeout), _MAX_TIMEOUT)
        az_auth_types = kwargs.pop("auth_types", kwargs.pop("mp_az_auth", None))
        if isinstance(az_auth_types, bool):
            self._az_auth_types = None
        elif isinstance(az_auth_types, str):
            self._az_auth_types = [az_auth_types]
        else:
            self._az_auth_types = az_auth_types
        self._az_tenant_id = kwargs.pop(
            "tenant_id", kwargs.pop("mp_az_tenant_id", None)
        )

        cluster = kwargs.pop("cluster", None)
        self.current_connection = connection_str or self.current_connection
        if not connection_str and not cluster:
            raise MsticpyParameterError(
                "Must specify either a connection string or a cluster name",
                parameter=["connection_str", "cluster"],
            )

        if cluster:
            self._current_config = self._lookup_cluster_settings(cluster)
            if not self._az_tenant_id:
                self._az_tenant_id = self._current_config.tenant_id
            logger.info(
                "Using cluster id: %s, retrieved url %s to build connection string",
                cluster,
                self.cluster_uri,
            )
            kusto_cs = self._get_connection_string_for_cluster(self._current_config)
            self.current_connection = cluster
        else:
            logger.info("Using connection string %s", connection_str)
            self.current_connection = connection_str
            kusto_cs = connection_str

        self.client = KustoClient(kusto_cs)
        proxies = kwargs.get("proxies", self._def_proxies)
        proxy_url = proxies.get("https") if proxies else None
        if proxy_url:
            logger.info(
                "Using proxy: %s",
                (
                    proxy_url
                    if "@" not in proxy_url  # don't log proxy credentials
                    else "****" + proxy_url.split("@")[-1]
                ),
            )
            self.client.set_proxy(proxy_url)
        self._connected = True

    def query(
        self, query: str, query_source: Optional[QuerySource] = None, **kwargs
    ) -> Union[pd.DataFrame, Any]:
        """
        Execute query string and return DataFrame of results.

        Parameters
        ----------
        query : str
            The query to execute
        query_source : QuerySource
            The query definition object

        Other Parameters
        ----------------
        database : str, Optional
            Supply or override the Kusto database name
        timeout : int
            Query timeout in seconds, default is 240 seconds (4 minutes)
            Maximum is 3600 seconds (1 hour).

        Returns
        -------
        Union[pd.DataFrame, results.ResultSet]
            A DataFrame (if successful) or
            the underlying provider result if an error.

        """
        data, result = self.query_with_results(
            query, query_source=query_source, **kwargs
        )
        return data if data is not None else result

    def query_with_results(  # type: ignore
        self, query: str, **kwargs
    ) -> Tuple[Optional[pd.DataFrame], Any]:
        """
        Return query results as a DataFrame and the result status.

        Parameters
        ----------
        query : str
            The query string

        Returns
        -------
        Tuple[Optional[pd.DataFrame], Any]
            DataFrame of results and the result status.

        Raises
        ------
        MsticpyNotConnectedError
            If there is no connection to the data source.
        MsticpyDataQueryError
            If no database is specified in the query or parameters
            and there is no default database.

        """
        if not self._connected:
            _raise_not_connected_error()
        query_source = kwargs.pop("query_source", None)

        if query_source and not self.query_usable(query_source):
            query_spec = self._get_cluster_spec_from_query_source(query_source)
            raise MsticpyDataQueryError(
                "Invalid query source - for this connection.",
                f"Connected cluster is: {self.cluster_uri} ({self.cluster_config_name})",
                "The cluster in the query definition is:",
                *[f"{name}: {value}" for name, value in query_spec.items()],
                title="Mismatched cluster for query.",
                help_uri=_HELP_URL,
            )

        database = self._get_query_database_name(query_source=query_source, **kwargs)
        data: Optional[pd.DataFrame] = None
        status = {"success": False}
        connection_props = ClientRequestProperties()
        connection_props.set_option(
            ClientRequestProperties.request_timeout_option_name,
            timedelta(seconds=kwargs.get("timeout", self._def_timeout)),
        )

        try:
            logger.info("Query executed query=%s, database=%s", query, database)
            response = self.client.execute(  # type: ignore[union-attr]
                database=database, query=query, properties=connection_props
            )
            data = dataframe_from_result_table(response.primary_results[0])
            status = _parse_query_status(response)
            logger.info("Query completed: %s", str(status))
        except KustoApiError as err:
            logger.error("Query failed: %s", err)
            _raise_kusto_error(err)
        except KustoServiceError as err:
            logger.error("Query failed: %s", err)
            _raise_unknown_query_error(err)
        return data, status

    def get_database_names(self) -> List[str]:
        """Get a list of database names from the connected cluster."""
        if self.client is None:
            _raise_not_connected_error()
        try:
            connection_props = ClientRequestProperties()
            connection_props.set_option(
                ClientRequestProperties.request_timeout_option_name,
                timedelta(seconds=self._def_timeout),
            )
            logger.info("Get database names cluster: %s", self.cluster_uri)
            response = self.client.execute_mgmt(  # type: ignore[union-attr]
                database="NetDefaultDB",
                query=".show databases",
                properties=connection_props,
            )

            # Convert the result to a DataFrame
            databases_df = dataframe_from_result_table(response.primary_results[0])
            return databases_df["DatabaseName"].tolist()
        except KustoServiceError as err:
            raise MsticpyDataQueryError(
                "Error getting database names",
                err,
                title="Kusto error",
                help_uri=_HELP_URL,
            ) from err

    def get_database_schema(
        self, database: Optional[str] = None
    ) -> Dict[str, Dict[str, str]]:
        """
        Get table names and schema from the connected cluster/database.

        Parameters
        ----------
        database : Optional[str]
            Name of the database to get schema for.
            The default is the last connected database.

        Returns
        -------
        Dict[str, Dict[str, str]]
            Dictionary of table names, each with a dictionary of
            column names and types.

        Raises
        ------
        ValueError :
            No database name specified or set as the default.
        MsticpyNotConnectedError :
            Not connected to a cluster.
        MsticpyDataQueryError :
            Error querying the cluster.

        """
        db_name = database or self._default_database
        if self.client is None:
            _raise_not_connected_error()
        if not db_name:
            raise ValueError("No database name specified")

        query = f".show database {db_name} schema"
        try:
            # Execute the query
            logger.info("Get database schema: %s", db_name)
            response = self.client.execute_mgmt(db_name, query)  # type: ignore[union-attr]
            # Convert the result to a DataFrame
            schema_dataframe = dataframe_from_result_table(response.primary_results[0])
        except KustoServiceError as err:
            raise MsticpyDataQueryError(
                "Error getting database schema",
                err,
                title="Kusto error",
                help_uri=_HELP_URL,
            ) from err

        return {
            str(table): {
                col_name: col_type.replace("System.", "")
                for col_name, col_type in cols[["ColumnName", "ColumnType"]].values
                if col_name is not None
            }
            for table, cols in schema_dataframe.groupby("TableName")
        }

    def _get_cluster_spec_from_query_source(
        self, query_source: QuerySource
    ) -> Dict[str, str]:
        """Return cluster details from query source."""
        return {
            QuerySourceFields.CLUSTER: query_source.metadata.get(
                QuerySourceFields.CLUSTER, "NA"
            ),
            QuerySourceFields.CLUSTERS: query_source.metadata.get(
                QuerySourceFields.CLUSTERS, "NA"
            ),
            QuerySourceFields.CLUSTER_GROUPS: query_source.metadata.get(
                QuerySourceFields.CLUSTER_GROUPS, "NA"
            ),
        }

    def _get_connection_string_for_cluster(
        self, cluster_config: KustoConfig
    ) -> KustoConnectionStringBuilder:
        """Return full cluster URI and credential for cluster name or URI."""
        auth_params = self._get_auth_params_from_config(cluster_config)
        connect_auth_types = self._az_auth_types or AzureCloudConfig().auth_methods
        if auth_params.method == "clientsecret":
            logger.info("Client secret specified in config - using client secret authn")
            if "clientsecret" not in connect_auth_types:
                connect_auth_types.insert(0, "clientsecret")
            credential = az_connect(
                auth_types=connect_auth_types, **(auth_params.params)
            )
        elif auth_params.method == "certificate":
            logger.info("Certificate specified in config - using certificate authn")
            connect_auth_types.insert(0, "certificate")
            credential = az_connect(
                auth_types=self._az_auth_types, **(auth_params.params)
            )
            return self._create_kusto_cert_connection_str(auth_params)
        else:
            logger.info("Using integrated authn")
            credential = az_connect(
                auth_types=self._az_auth_types, **(auth_params.params)
            )
        logger.info("Credentials obtained %s", type(credential.modern).__name__)
        token = credential.modern.get_token(get_default_resource_name(auth_params.uri))
        logger.info("Token obtained for %s", auth_params.uri)
        return KustoConnectionStringBuilder.with_aad_user_token_authentication(
            connection_string=auth_params.uri,
            user_token=token.token,
        )

    def _create_kql_cert_connection_str(
        self, auth_params: AuthParams
    ) -> KustoConnectionStringBuilder:
        logger.info("Creating KQL connection string for certificate authentication")
        if not self._az_tenant_id:
            raise ValueError(
                "Azure tenant ID must be set in config or connect parameter",
                "to use certificate authentication",
            )
        cert_bytes = base64.b64decode(auth_params.params["certificate"])
        (
            private_key,
            certificate,
            _,
        ) = pkcs12.load_key_and_certificates(data=cert_bytes, password=None)
        if private_key is None or certificate is None:
            raise ValueError(
                f"Could not load certificate for cluster {self.cluster_uri}"
            )
        private_cert = private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.TraditionalOpenSSL,
            encryption_algorithm=serialization.NoEncryption(),
        )
        public_cert = certificate.public_bytes(encoding=serialization.Encoding.PEM)
        thumbprint = certificate.fingerprint(hashes.SHA256())
        return KustoConnectionStringBuilder.with_aad_application_certificate_sni_authentication(
            connection_string=self.cluster_uri,
            aad_app_id=auth_params.params["client_id"],
            private_certificate=private_cert.decode("utf-8"),
            public_certificate=public_cert.decode("utf-8"),
            thumbprint=thumbprint.hex().upper(),
            authority_id=self._az_tenant_id,
        )

    def _get_auth_params_from_config(self, cluster_config: KustoConfig) -> AuthParams:
        """Get authentication parameters for cluster from KustoConfig values."""
        method = "integrated"
        auth_params_dict = {}
        if KFields.CLIENT_SEC in cluster_config and KFields.CLIENT_ID in cluster_config:
            method = "clientsecret"
            auth_params_dict["client_id"] = cluster_config.ClientId
            auth_params_dict["client_secret"] = cluster_config.ClientSecret
            logger.info(
                "Using client secret authentication because client_secret in config"
            )
        elif (
            KFields.CERTIFICATE in cluster_config
            and KFields.CLIENT_ID in cluster_config
        ):
            method = "certificate"
            auth_params_dict["client_id"] = cluster_config.ClientId
            auth_params_dict["certificate"] = cluster_config.Certificate
            logger.info(
                "Using client secret authentication because client_secret in config"
            )
        elif KFields.INTEG_AUTH in cluster_config:
            logger.info("Using integrated auth.")
        auth_params_dict["tenant_id"] = cluster_config.tenant_id
        return AuthParams(method, auth_params_dict, cluster_config.cluster)

    def _lookup_cluster_settings(self, cluster: str) -> KustoConfig:
        """Return cluster URI from config if cluster name is passed."""
        cluster_key = cluster.casefold().strip()
        if cluster_key in self._kusto_settings["url"]:
            return self._kusto_settings["url"][cluster_key]
        if cluster_key in self._kusto_settings["name"]:
            return self._kusto_settings["name"][cluster_key]
        if cluster_key in self._kusto_settings["id"]:
            return self._kusto_settings["id"][cluster_key]
        if cluster_key.startswith("https://"):
            return KustoConfig(
                cluster=cluster,
                name=cluster.replace("https://", "").split(".")[0],
                alias="no_config_found",
                path="",
                tenant_id=self._az_tenant_id,
            )

        raise MsticpyDataQueryError(
            f"Cluster '{cluster}' not found in msticpyconfig.yaml",
            "or is not in the correct format for a a cluster URI",
            "The cluster must be a key, cluster short name of an entry defined",
            "in the 'KustoClusters' section of the config file,",
            "or it must be a valid cluster URI.",
            title="Unusable cluster identifier",
            help_uri=_HELP_URL,
        )

    def _get_query_database_name(
        self, query_source: Optional[QuerySource] = None, **kwargs
    ) -> str:
        """Get the database name from query source or kwargs."""
        if database := kwargs.get("database"):
            logger.info("Using database %s from parameter.", database)
            return database
        if query_source:
            return self._get_db_from_query_source(query_source)
        # check if database is specified in the current config
        if isinstance(self._current_config, KustoConfig):
            database = self._current_config.default_db
            if database:
                logger.info("Using database %s from current config.", database)
                return database
        if self._default_database:
            logger.info("Using database %s from _default_database.", database)
            return self._default_database
        _raise_no_db_error()
        return ""  # pragma: no cover

    def query_usable(self, query_source) -> bool:
        """Return True if query source is valid for current cluster."""
        if not query_source or not isinstance(query_source, QuerySource):
            return False
        if not self._current_config:
            # This probably won't work but we can just try the query
            # and see if it works
            return True
        # Check if query source has cluster information
        if not (
            query_source.metadata.keys()
            & {
                QuerySourceFields.CLUSTER_GROUPS,
                QuerySourceFields.CLUSTERS,
                QuerySourceFields.CLUSTER,
            }
        ):
            if self._strict_query_match:
                return False
            logger.info(
                (
                    "Query source %s has no cluster information. "
                    "The query may fail on the current cluster."
                ),
                query_source.name,
            )
            return True
        # Check for matches on cluster groups or cluster id
        result = False
        result |= self._cluster_groups_match(query_source)
        if result:
            return result
        result |= self._cluster_id_matches(query_source)
        return result

    def _cluster_groups_match(self, query_source: QuerySource) -> bool:
        """Return True if query source cluster group is valid for current cluster."""
        source_cluster_groups = query_source.metadata.get(
            QuerySourceFields.CLUSTER_GROUPS, []
        )
        if (
            source_cluster_groups
            and self._current_config.cluster_groups  # type: ignore[union-attr]
        ):
            driver_groups = {
                group.casefold()
                for group in self._current_config.cluster_groups  # type: ignore[union-attr]
            }
            query_groups = {group.casefold() for group in source_cluster_groups}
            return bool(driver_groups.intersection(query_groups))
        return False

    def _cluster_id_matches(self, query_source: QuerySource) -> bool:
        """Return True if query source cluster is valid for current cluster."""
        # Get different representations of the cluster name
        result = False
        cluster_ids = {
            self.cluster_uri.casefold(),
            self.cluster_name.casefold(),
            self.cluster_config_name.casefold(),
        }
        source_clusters = query_source.metadata.get(QuerySourceFields.CLUSTERS, [])
        if source_clusters:
            query_source_clusters = {cluster.casefold() for cluster in source_clusters}
            result |= bool(cluster_ids.intersection(query_source_clusters))
            if result:
                return result
        source_cluster = query_source.metadata.get(QuerySourceFields.CLUSTER)
        if source_cluster:
            result |= source_cluster.casefold() in cluster_ids
        return result

    @staticmethod
    def _get_db_from_query_source(query_source: QuerySource) -> str:
        """Get the database name from query source metadata."""
        if database := query_source.metadata.get("database"):
            return database
        data_families = query_source.metadata.get("data_families")
        if not data_families:
            logger.info("Could not find database name in query source metadata.")
            _raise_no_db_error(query_source)

        if "." in data_families[0]:  # type: ignore
            _, database = data_families[0].split(".", maxsplit=1)  # type: ignore
        else:
            # Not expected but we can still use a DB value with no dot
            database = data_families[0]  # type: ignore
        logger.info("Using database %s from query source metadata.", database)
        return database


def _get_kusto_settings() -> Dict[str, Dict[str, KustoConfig]]:
    """Return a dictionary of Kusto cluster settings from msticpyconfig.yaml."""
    kusto_config = {
        kusto_entry: kusto_config
        for kusto_entry, kusto_config in get_config("DataProviders", {}).items()
        if kusto_entry.startswith("Kusto")
    }
    kusto_clusters = {}
    # handle legacy configuration
    for config_id, cluster_conf in kusto_config.items():
        name = config_id.replace("Kusto-", "")
        kusto_clusters[name] = cluster_conf
        kusto_clusters[name]["path"] = f"DataProviders.{config_id}"

    kusto_new_conf = {
        config_id: {**cluster_conf, "path": f"KustoClusters.{config_id}"}
        for config_id, cluster_conf in get_config("KustoClusters", {}).items()
    }
    defaults: Dict[str, Any] = kusto_new_conf.pop(KFields.DEFAULTS, {}).get(
        KFields.ARGS, {}  # type: ignore[assignment]
    )
    kusto_clusters.update(kusto_new_conf)

    cluster_by_url = _create_cluster_config(
        kusto_clusters=kusto_clusters, defaults=defaults
    )
    return {
        "url": cluster_by_url,
        "id": {conf.alias.casefold(): conf for conf in cluster_by_url.values()},
        "name": {conf.name.casefold(): conf for conf in cluster_by_url.values()},
    }


def _create_cluster_config(
    kusto_clusters: Dict[str, Any], defaults: Dict[str, Any]
) -> Dict[str, KustoConfig]:
    """Return a dictionary of Kusto cluster settings from msticpyconfig.yaml."""
    return {
        config[KFields.ARGS]
        .get(KFields.CLUSTER)
        .casefold(): KustoConfig(
            tenant_id=_setting_or_default(
                config[KFields.ARGS], KFields.TENANT_ID, defaults
            ),
            integrated_auth=_setting_or_default(
                config[KFields.ARGS], KFields.INTEG_AUTH, defaults
            )
            or False,
            args=_create_protected_args(
                _section_or_default(config[KFields.ARGS], defaults), config["path"]
            ),
            cluster=config[KFields.ARGS].get(KFields.CLUSTER),
            alias=name,
            name=get_cluster_name(config[KFields.ARGS].get(KFields.CLUSTER)),
            path=config["path"],
            cluster_groups=config.get(KFields.CLUSTER_GROUPS),
        )
        for name, config in kusto_clusters.items()
    }


def _setting_or_default(settings: Dict[str, Any], name: str, default: Dict[str, Any]):
    """Return a setting from the settings dictionary or the default."""
    return settings.get(name, default.get(name))


def _section_or_default(settings: Dict[str, Any], default: Dict[str, Any]):
    """Return a combined dictionary from the settings dictionary or the default."""
    return {
        key: settings.get(key, default.get(key))
        for key in (settings.keys() | default.keys())
    }


def _create_protected_args(args: Dict[str, Any], path: str) -> ProviderArgs:
    """Return a dictionary of protected settings for Kusto args config."""
    args_dict = {
        key_name: (
            partial(
                get_protected_setting, config_path=f"{path}.Args", setting_name=key_name
            )
            if isinstance(value, dict)
            and (value.get("EnvironmentVar") or value.get("KeyVault"))
            else value
        )
        for key_name, value in args.items()
    }
    return ProviderArgs(**args_dict)


def get_cluster_name(cluster_uri):
    """Return the cluster name from the cluster uri."""
    return cluster_uri.replace("https://", "").split(".")[0]


def _parse_query_status(response: KustoResponseDataSet) -> Dict[str, Any]:
    """Parse the query status from the Kusto response."""
    try:
        query_info_idx = response.tables_names.index("QueryCompletionInformation")
    except ValueError:
        return {
            "status": "Failed",
            "message": "QueryCompletionInformation not found in response",
        }

    df_status = dataframe_from_result_table(response.tables[query_info_idx])
    results = df_status[["EventTypeName", "Payload"]].to_dict(orient="records")
    return {
        row.get("EventTypeName", "Unknown_field"): json.loads(
            row.get("Payload", "No Payload")
        )
        for row in results
    }


def _raise_kusto_error(error):
    """Raise a Kusto error."""
    if isinstance(error, KustoApiError):
        raise MsticpyDataQueryError(
            error.error.description,
            f"error code: {error.error.code}",
            title=error.error.message,
            help_uri=_HELP_URL,
        ) from error


def _raise_no_db_error(query_source=None):
    """Raise an error if no database is specified."""
    if query_source:
        messages = (
            "No database found in the query definition",
            (
                "Correct the query definition or use the 'database' parameter"
                "or set the default database when connecting to the cluster."
            ),
        )
    else:
        messages = (
            (
                "No database specified. Use the 'database' parameter or set the "
                "default database when connecting to the cluster."
            ),
        )
    raise MsticpyDataQueryError(
        *messages,
        title="No database specified",
        help_uri=_HELP_URL,
    )


def _raise_not_connected_error():
    """Raise an error if not connected to a cluster."""
    raise MsticpyNotConnectedError(
        "Please connect to the cluster before executing a query.",
        title="Not connected to cluster",
        help_uri=_HELP_URL,
    )


def _raise_unknown_query_error(err):
    """Raise an error if unknown exception raised."""
    raise MsticpyDataQueryError(
        "Unknown exception when executing query.",
        f"Exception type: {type(err)}",
        *err.args,
        title="Unknown exception during query execution",
        help_uri=_HELP_URL,
    ) from err



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Azure monitor/Log Analytics KQL Driver class.

See Also
--------
Azure SDK code: https://github.com/Azure/azure-sdk-for-python/tree/main/sdk/monitor

Azure SDK docs: https://learn.microsoft.com/python/api/overview/
azure/monitor-query-readme?view=azure-python

"""
import contextlib
import logging
import warnings
from datetime import datetime
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union, cast

import httpx
import pandas as pd
from azure.core.exceptions import HttpResponseError
from azure.core.pipeline.policies import UserAgentPolicy
from packaging.version import parse as parse_version

from ..._version import VERSION
from ...auth.azure_auth import AzureCloudConfig, az_connect
from ...common.exceptions import (
    MsticpyDataQueryError,
    MsticpyKqlConnectionError,
    MsticpyMissingDependencyError,
    MsticpyNoDataSourceError,
    MsticpyNotConnectedError,
)
from ...common.provider_settings import get_protected_setting
from ...common.settings import get_http_proxies, get_http_timeout
from ...common.timespan import TimeSpan
from ...common.utility import export, mp_ua_header
from ...common.wsconfig import WorkspaceConfig
from ..core.query_defns import DataEnvironment
from .driver_base import DriverBase, DriverProps, QuerySource

logger = logging.getLogger(__name__)

# pylint: disable=ungrouped-imports
try:
    from azure.monitor.query import (
        LogsQueryClient,
        LogsQueryPartialResult,
        LogsQueryResult,
    )
    from azure.monitor.query import __version__ as az_monitor_version
except ImportError as imp_err:
    raise MsticpyMissingDependencyError(
        "Cannot use this feature without Azure monitor client installed",
        title="Error importing azure.monitor.query",
        packages="azure-monitor-query",
    ) from imp_err

__version__ = VERSION
__author__ = "Ian Hellen"


_KQL_CLOUD_MAP = {"global": "public", "cn": "china", "usgov": "government"}

_HELP_URL = (
    "https://msticpy.readthedocs.io/en/latest/data_acquisition/DataProv-MSSentinel.html"
)
# pylint: disable=too-many-instance-attributes


@export
class AzureMonitorDriver(DriverBase):
    """KqlDriver class to execute kql queries."""

    _DEFAULT_TIMEOUT = 300

    def __init__(self, connection_str: Optional[str] = None, **kwargs):
        """
        Instantiate KqlDriver and optionally connect.

        Parameters
        ----------
        connection_str : str, optional
            Connection string

        Other Parameters
        ----------------
        debug : bool
            print out additional diagnostic information.
        timeout : int (seconds)
            Specify a timeout for queries. Default is 300 seconds.
            (can be set here or in connect and overridden in query methods)
        proxies : Dict[str, str]
            Proxy settings for log analytics queries.
            Dictionary format is {protocol: proxy_url}
            Where protocol is https, http, etc. and proxy_url can contain
            optional authentication information in the format
            "https://username:password@proxy_host:port"
            If you have a proxy configuration in msticpyconfig.yaml and
            you do not want to use it, set this to an empty dictionary.
            (can be overridden in connect method)

        """
        if kwargs.get("debug", False):
            logger.setLevel(logging.DEBUG)
        super().__init__(**kwargs)

        self._schema: Dict[str, Any] = {}
        self.set_driver_property(
            DriverProps.FORMATTERS,
            {"datetime": self._format_datetime, "list": self._format_list},
        )
        self._loaded = True
        self._ua_policy = UserAgentPolicy(user_agent=mp_ua_header()["UserAgent"])
        self._def_timeout = kwargs.get(
            "timeout", kwargs.get("server_timeout", self._DEFAULT_TIMEOUT)
        )
        self._def_proxies = kwargs.get("proxies", get_http_proxies())
        self._query_client: Optional[LogsQueryClient] = None
        self._az_tenant_id: Optional[str] = None
        self._ws_config: Optional[WorkspaceConfig] = None
        self._ws_name: Optional[str] = None
        self._workspace_id: Optional[str] = None
        self._workspace_ids: List[str] = []
        self._def_connection_str: Optional[str] = connection_str
        self._connect_auth_types: Optional[List[str]] = None
        self.add_query_filter(
            "data_environments", ("MSSentinel", "LogAnalytics", "AzureSentinel")
        )
        self.set_driver_property(
            DriverProps.EFFECTIVE_ENV, DataEnvironment.MSSentinel.name
        )
        self.set_driver_property(DriverProps.SUPPORTS_THREADING, value=True)
        self.set_driver_property(
            DriverProps.MAX_PARALLEL, value=kwargs.get("max_threads", 4)
        )
        self.az_cloud_config = AzureCloudConfig()
        logger.info(
            "AzureMonitorDriver loaded. connect_str  %s, kwargs: %s",
            connection_str,
            kwargs,
        )

    @property
    def url_endpoint(self) -> str:
        """Return the current URL endpoint for Azure Monitor."""
        base_url = self.az_cloud_config.log_analytics_uri
        # post v1.1.0 of azure-monitor-query, the API version requires a 'v1' suffix
        if parse_version(az_monitor_version) > parse_version("1.1.0"):
            return f"{base_url}v1"
        return base_url

    @property
    def current_connection(self) -> str:
        """Return the current connection name."""
        connection = self._ws_name
        if (
            not connection
            and self._ws_config
            and WorkspaceConfig.CONF_WS_NAME_KEY in self._ws_config
        ):
            connection = self._ws_config[WorkspaceConfig.CONF_WS_NAME_KEY]
        return (
            connection
            or self._def_connection_str
            or self._workspace_id
            or next(iter(self._workspace_ids), "")
            or "AzureMonitor"
        )

    @current_connection.setter
    def current_connection(self, value: str):
        """Allow attrib to be set but ignore."""
        del value

    def connect(self, connection_str: Optional[str] = None, **kwargs):
        """
        Connect to data source.

        Parameters
        ----------
        connection_str : Union[str, WorkspaceConfig, None]
            Connection string or WorkspaceConfig for the Sentinel Workspace.

        Other Parameters
        ----------------
        auth_types: Iterable [str]
            Authentication (credential) types to use. By default the
            values configured in msticpyconfig.yaml are used. If not set,
            it will use the msticpy defaults.
        mp_az_auth : Union[bool, str, list, None], optional
            Deprecated parameter directing driver to use MSTICPy Azure authentication.
            Values can be:
            True or "default": use the settings in msticpyconfig.yaml 'Azure' section
            str: single auth method name
            List[str]: list of acceptable auth methods from
            Use `auth_types` parameter instead.
        tenant_id: str, optional
            Optional parameter specifying a Tenant ID for use by MSTICPy Azure
            authentication. By default, the tenant_id for the workspace.
        workspace : str, optional
            Alternative to supplying a WorkspaceConfig object as the connection_str
            parameter. Giving a workspace name will fetch the workspace
            settings from msticpyconfig.yaml.
        workspaces : Iterable[str], optional
            List of workspaces to run the queries against, each workspace name
            must have an entry in msticpyconfig.yaml
        workspace_ids: Iterable[str], optional
            List of workspace IDs to run the queries against. Must be supplied
            along with a `tenant_id`.
        timeout : int (seconds)
            Specify a timeout for queries. Default is 300 seconds.
            (can be overridden query method)
        proxies : Dict[str, str]
            Proxy settings for log analytics queries.
            Dictionary format is {protocol: proxy_url}
            Where protocol is https, http, etc. and proxy_url can contain
            optional authentication information in the format
            "https://username:password@proxy_host:port"
            If you have a proxy configuration in msticpyconfig.yaml and
            you do not want to use it, set this to an empty dictionary.

        Notes
        -----
        When using the `workspaces` or `workspace_ids` parameters, some
        functionality will be reduced - e.g. no schema will be available for
        the workspaces. As an alternative to using multiple workspaces here
        you can create multiple workspace connections

        """
        self._connected = False
        self._query_client = self._create_query_client(connection_str, **kwargs)

        # get the schema
        self._schema = self._get_schema()
        self._connected = True
        print("connected")

        return self._connected

    # pylint: disable=too-many-branches

    @property
    def schema(self) -> Dict[str, Dict]:
        """
        Return current data schema of connection.

        Returns
        -------
        Dict[str, Dict]
            Data schema of current connection.

        """
        return self._schema

    def query(
        self, query: str, query_source: Optional[QuerySource] = None, **kwargs
    ) -> Union[pd.DataFrame, Any]:
        """
        Execute query string and return DataFrame of results.

        Parameters
        ----------
        query : str
            The query to execute
        query_source : Optional[QuerySource]
            The query definition object

        Other Parameters
        ----------------
        timeout : int (seconds)
            Specify a timeout for the query. Default is 300 seconds.
        fail_on_partial: bool
            Fail the query if only partial results are returned.

        Returns
        -------
        Union[pd.DataFrame, results.ResultSet]
            A DataFrame (if successful) or
            the underlying provider result if an error.

        """
        if not self._connected or self._query_client is None:
            raise MsticpyNotConnectedError(
                "Please run connect() to connect to the workspace",
                "before running a query.",
                title="Workspace not connected.",
                help_uri=_HELP_URL,
            )
        if query_source:
            self._check_table_exists(query_source)
        data, result = self.query_with_results(query, **kwargs)
        return data if data is not None else result

    # pylint: disable=too-many-branches
    def query_with_results(
        self, query: str, **kwargs
    ) -> Tuple[pd.DataFrame, Dict[str, Any]]:
        """
        Execute query string and return DataFrame of results.

        Parameters
        ----------
        query : str
            The kql query to execute

        Returns
        -------
        Tuple[pd.DataFrame, Dict[str, Any]]
            A DataFrame (if successful) and
            Query status dictionary.

        """
        if not self._connected or self._query_client is None:
            raise MsticpyNotConnectedError(
                "Please run connect() to connect to the workspace",
                "before running a query.",
                title="Workspace not connected.",
                help_uri=_HELP_URL,
            )
        time_span_value = self._get_time_span_value(**kwargs)
        fail_on_partial = kwargs.get(
            "fail_if_partial", kwargs.get("fail_on_partial", False)
        )
        server_timeout = kwargs.pop("timeout", self._def_timeout)

        workspace_id = next(iter(self._workspace_ids), None) or self._workspace_id
        additional_workspaces = self._workspace_ids[1:] if self._workspace_ids else None
        logger.info("Query to run %s", query)
        logger.info(
            "Workspaces %s", ",".join(self._workspace_ids) or self._workspace_id
        )
        logger.info(
            "Time span %s - %s",
            str(time_span_value[0]) if time_span_value else "none",
            str(time_span_value[1]) if time_span_value else "none",
        )
        logger.info("Timeout %s", server_timeout)
        try:
            result = self._query_client.query_workspace(
                workspace_id=workspace_id,  # type: ignore[arg-type]
                query=query,
                timespan=time_span_value,  # type: ignore[arg-type]
                server_timeout=server_timeout,
                additional_workspaces=additional_workspaces,
            )
        except HttpResponseError as http_err:
            result = None
            self._raise_query_failure(query, http_err)
        # We might get an unknown exception type from azure.monitor.query
        except Exception as unknown_err:  # pylint: disable=broad-except
            result = None
            self._raise_unknown_error(unknown_err)
        result = cast(LogsQueryResult, result)
        status = self._get_query_status(result)
        logger.info("query status %s", repr(status))

        if isinstance(result, LogsQueryPartialResult):
            if fail_on_partial:
                raise MsticpyDataQueryError(
                    "Partial results returned. This may indicate a query timeout.",
                    title="Partial results returned",
                    help_uri=_HELP_URL,
                )
            warnings.warn(
                "Partial results returned. This may indicate a query timeout.",
                RuntimeWarning,
            )
            table = result.partial_data[0]  # type: ignore[attr-defined]
        else:
            table = result.tables[0]  # type: ignore[attr-defined]
        data_frame = pd.DataFrame(table.rows, columns=table.columns)
        logger.info("Dataframe returned with %d rows", len(data_frame))
        return data_frame, status

    def _create_query_client(self, connection_str, **kwargs):
        """Create the query client."""
        az_auth_types = kwargs.pop("auth_types", kwargs.get("mp_az_auth"))
        if isinstance(az_auth_types, bool):
            az_auth_types = None
        if isinstance(az_auth_types, str):
            az_auth_types = [az_auth_types]
        self._connect_auth_types = az_auth_types

        self._def_timeout = kwargs.pop("timeout", self._DEFAULT_TIMEOUT)
        self._def_proxies = kwargs.pop("proxies", self._def_proxies)
        self._get_workspaces(connection_str, **kwargs)

        # check for additional Args in settings but allow kwargs to override
        connect_args = self._get_workspace_settings_args()
        connect_args.update(kwargs)
        connect_args.update(
            {"auth_methods": az_auth_types, "tenant_id": self._az_tenant_id}
        )
        credentials = az_connect(**connect_args)
        logger.info(
            "Created query client. Auth type: %s, Url: %s, Proxies: %s",
            type(credentials.modern) if credentials else "None",
            self.url_endpoint,
            kwargs.get("proxies", self._def_proxies),
        )
        return LogsQueryClient(
            credential=credentials.modern,
            endpoint=self.url_endpoint,
            proxies=kwargs.get("proxies", self._def_proxies),
        )

    def _get_workspace_settings_args(self) -> Dict[str, Any]:
        """Return any Args settings for the current workspace."""
        if not self._ws_config or not self._ws_config.settings_path:
            return {}
        args_path = f"{self._ws_config.settings_path}.Args"
        args_settings = self._ws_config.settings.get("Args", {})
        return {
            name: get_protected_setting(args_path, name)
            for name in args_settings.keys()
        }

    def _get_workspaces(self, connection_str: Optional[str] = None, **kwargs):
        """Get workspace or workspaces to connect to."""
        self._az_tenant_id = kwargs.get("tenant_id", kwargs.get("mp_az_tenant_id"))
        # multiple workspace IDs
        if workspaces := kwargs.pop("workspaces", None):
            self._get_workspaces_by_name(workspaces)
            return
        if workspace_ids := kwargs.pop("workspace_ids", None):
            self._get_workspaces_by_id(workspace_ids)
            return

        # standard - single-workspace configuration
        workspace_name = kwargs.get("workspace")
        ws_config: Optional[WorkspaceConfig] = None
        connection_str = connection_str or self._def_connection_str
        if workspace_name or connection_str is None:
            ws_config = WorkspaceConfig(workspace=workspace_name)  # type: ignore
            logger.info(
                "WorkspaceConfig created from workspace name %s", workspace_name
            )
        elif isinstance(connection_str, str):
            self._def_connection_str = connection_str
            with contextlib.suppress(ValueError):
                ws_config = WorkspaceConfig.from_connection_string(connection_str)
                logger.info(
                    "WorkspaceConfig created from connection_str %s", connection_str
                )
        elif isinstance(connection_str, WorkspaceConfig):
            logger.info("WorkspaceConfig as parameter %s", connection_str.workspace_id)
            ws_config = connection_str

        if not ws_config:
            logger.warning("No workspace set")
            raise MsticpyKqlConnectionError(
                "A workspace name, config or connection string is needed"
                " to connect to a workspace.",
                title="No connection details",
                help_uri=_HELP_URL,
            )
        if ws_config.workspace_id is None or ws_config.tenant_id is None:
            logger.warning("Unable to get workspace ID or tenant ID")
            raise MsticpyKqlConnectionError(
                "The workspace config or connection string did not have"
                "the required parameters to connect to a workspace.",
                "At least a workspace ID and tenant ID are required.",
                title="No connection details",
                help_uri=_HELP_URL,
            )
        self._ws_config = ws_config
        self._ws_name = workspace_name or ws_config.workspace_id
        if not self._az_tenant_id and WorkspaceConfig.CONF_TENANT_ID in ws_config:
            self._az_tenant_id = ws_config[WorkspaceConfig.CONF_TENANT_ID]
        self._workspace_id = ws_config[WorkspaceConfig.CONF_WS_ID]

    def _get_workspaces_by_id(self, workspace_ids):
        if not self._az_tenant_id:
            raise MsticpyKqlConnectionError(
                "You must supply a tenant_id with the workspace_ids parameter",
                title="No tenant_id supplied.",
                help_uri=_HELP_URL,
            )
        self._workspace_ids = workspace_ids
        logger.info(
            "%d configured workspaces: %s",
            len(self._workspace_ids),
            ", ".join(self._workspace_ids),
        )

    def _get_workspaces_by_name(self, workspaces):
        workspace_configs = {
            WorkspaceConfig(workspace)[WorkspaceConfig.CONF_WS_ID]: WorkspaceConfig(
                workspace
            )[WorkspaceConfig.CONF_TENANT_ID]
            for workspace in workspaces
        }
        if len(set(workspace_configs.values())) > 1:
            raise ValueError("All workspaces must have the same tenant ID.")
        self._az_tenant_id = next(iter(workspace_configs.values()))
        self._workspace_ids = list(set(workspace_configs))
        logger.info(
            "%d configured workspaces: %s",
            len(self._workspace_ids),
            ", ".join(self._workspace_ids),
        )

    def _get_time_span_value(self, **kwargs):
        """Return the timespan for the query API call."""
        default_time_params = kwargs.get("default_time_params", False)
        time_params = kwargs.get("time_span", {})
        start = time_params.get("start")
        end = time_params.get("end")
        if default_time_params or start is None or end is None:
            time_span_value = None
            logger.info("No time parameters supplied.")
        else:
            time_span = TimeSpan(
                start=time_params["start"],
                end=time_params["end"],
            )
            # Azure Monitor API expects datetime objects, so
            # convert to datetimes if we have pd.Timestamps
            t_start = (
                time_span.start.to_pydatetime(warn=False)
                if isinstance(time_span.start, pd.Timestamp)
                else time_span.start
            )
            t_end = (
                time_span.end.to_pydatetime(warn=False)
                if isinstance(time_span.end, pd.Timestamp)
                else time_span.end
            )
            time_span_value = t_start, t_end
            logger.info("Time parameters set %s", str(time_span))
        return time_span_value

    def _check_table_exists(self, query_source):
        """Check that query table is in the workspace schema."""
        if not self.schema:
            return
        try:
            table = query_source.params.get("table", {}).get("default")
        except KeyError:
            table = None
        if table:
            if " " in table.strip():
                table = table.strip().split(" ")[0]
            if table not in self.schema:
                raise MsticpyNoDataSourceError(
                    f"The table {table} for this query is not in your workspace",
                    "or database schema. Please check your query.",
                    title=f"{table} not found.",
                    help_uri=_HELP_URL,
                )

    @staticmethod
    def _get_query_status(result) -> Dict[str, Any]:
        if isinstance(result, LogsQueryResult):
            return {
                "status": result.status.name,
                "tables": len(result.tables),
            }
        if isinstance(result, LogsQueryPartialResult):
            return {
                "status": result.status.name,
                "tables": len(result.partial_data),
            }
        return {"status": "unknown failure", "tables": 0, "result": result}

    def _get_schema(self) -> Dict[str, Dict]:
        """Return the workspace schema."""
        if not self._ws_config:
            logger.info("No workspace config - cannot get schema")
            return {}
        mgmt_endpoint = self.az_cloud_config.resource_manager

        url_tables = (
            "{endpoint}subscriptions/{sub_id}/resourcegroups/"
            "{res_group}/providers/Microsoft.OperationalInsights/workspaces/"
            "{ws_name}/tables?api-version=2021-12-01-preview"
        )
        try:
            ws_name = self._ws_config.workspace_name
        except AttributeError:
            ws_name = self._ws_config.workspace_key

        if not ws_name or ws_name == "Default":
            logger.info("No workspace name - cannot get schema")
            return {}
        try:
            fmt_url = url_tables.format(
                endpoint=mgmt_endpoint,
                sub_id=self._ws_config.subscription_id,
                res_group=self._ws_config.resource_group,
                ws_name=self._ws_config.workspace_name,
            )
        except AttributeError:
            logger.info("Not all workspace config available - cannot get schema")
            return {}

        credentials = az_connect(
            auth_methods=self._connect_auth_types, tenant_id=self._az_tenant_id
        )
        token = credentials.modern.get_token(f"{mgmt_endpoint}/.default")
        headers = {"Authorization": f"Bearer {token.token}", **mp_ua_header()}
        logger.info("Schema request to %s", fmt_url)
        response = httpx.get(
            fmt_url,
            headers=headers,
            timeout=get_http_timeout(),
            proxies=self._def_proxies or {},
        )
        if response.status_code != 200:
            logger.info("Schema request failed. Status code: %d", response.status_code)
            return {}
        tables = response.json()
        logger.info(
            "Schema retrieved from workspace. %d tables found.",
            len(tables.get("value", 0)),
        )
        return _schema_format_tables(tables)

    @staticmethod
    def _format_datetime(date_time: datetime) -> str:
        """Return datetime-formatted string."""
        return date_time.isoformat(sep="T") + "Z"

    @staticmethod
    def _format_list(param_list: Iterable[Any]):
        """Return formatted list parameter."""
        fmt_list = []
        for item in param_list:
            if isinstance(item, str):
                fmt_list.append(f"'{item}'")
            else:
                fmt_list.append(f"{item}")
        return ", ".join(fmt_list)

    @staticmethod
    def _raise_query_failure(query, http_err):
        """Raise query failure exception."""
        err_contents = []
        if hasattr(http_err, "message"):
            err_contents = http_err.message.split("\n")
        if not err_contents:
            err_contents = ["Unknown query error"]

        err_contents.append(f"Query:\n{query}")
        raise MsticpyDataQueryError(
            *err_contents, title="Query Failure", help_uri=_HELP_URL
        ) from http_err

    @staticmethod
    def _raise_unknown_error(exception):
        """Raise an unknown exception."""
        raise MsticpyDataQueryError(
            "An unknown exception was returned by the service",
            *exception.args,
            f"Full exception:\n{exception}",
            title="connection failed",
            help_uri=_HELP_URL,
        ) from exception


def _schema_format_tables(
    ws_tables: Dict[str, Iterable[Dict[str, Any]]]
) -> Dict[str, Dict[str, str]]:
    """Return a sorted dictionary of table names and column names/types."""
    table_schema = {
        table["name"]: _schema_format_columns(table["properties"]["schema"])
        for table in ws_tables["value"]
    }
    return dict(sorted(table_schema.items()))


def _schema_format_columns(table_schema: Dict[str, Any]) -> Dict[str, str]:
    """Return a sorted dictionary of column names and types."""
    columns = {
        col["name"]: col["type"] for col in table_schema.get("standardColumns", {})
    }
    for col in table_schema.get("customColumns", []):
        columns[col["name"]] = col["type"]
    return dict(sorted(columns.items()))
